from time import time
from unittest.mock import create_autospec, MagicMock

import pytest
from flask_pyoidc.auth_response_handler import AuthResponseHandler, AuthResponseUnexpectedStateError, \
    InvalidIdTokenError, AuthResponseErrorResponseError, AuthResponseMismatchingSubjectError
from flask_pyoidc.provider_configuration import ProviderConfiguration, ProviderMetadata, ClientMetadata
from flask_pyoidc.pyoidc_facade import PyoidcFacade
from oic.oic import AuthorizationResponse, AccessTokenResponse, IdToken, TokenErrorResponse, \
    OpenIDSchema, AuthorizationErrorResponse, AuthorizationRequest


def _create_id_token(issuer, client_id, nonce):
    id_token = IdToken(**{'iss': issuer, 'sub': 'test_sub', 'aud': client_id, 'nonce': nonce, 'exp': time() + 60})
    id_token.jws_header = {'alg': 'RS256'}
    return id_token


class TestAuthResponseHandler:
    ISSUER = 'https://issuer.example.com'
    CLIENT_ID = 'client1'
    AUTH_REQUEST = AuthorizationRequest(**{'state': 'test_state', 'nonce': 'test_nonce'})
    AUTH_RESPONSE = AuthorizationResponse(**{'code': 'test_auth_code', 'state': AUTH_REQUEST['state']})
    TOKEN_RESPONSE = AccessTokenResponse(**{
        'access_token': 'test_token',
        'expires_in': 3600,
        'id_token': _create_id_token(ISSUER, CLIENT_ID, AUTH_REQUEST['nonce']),
        'id_token_jwt': 'test_id_token_jwt',
        'refresh_token': 'test_refresh_token'
    })
    USERINFO_RESPONSE = OpenIDSchema(**{'sub': 'test_sub'})
    ERROR_RESPONSE = {'error': 'test_error', 'error_description': 'something went wrong'}

    @pytest.fixture
    def client_mock(self):
        return create_autospec(PyoidcFacade, True, True)

    def test_should_detect_state_mismatch(self, client_mock):
        auth_request = {'state': 'other_state', 'nonce': self.AUTH_REQUEST['nonce']}
        with pytest.raises(AuthResponseUnexpectedStateError):
            AuthResponseHandler(client_mock).process_auth_response(self.AUTH_RESPONSE, auth_request, {})

    def test_should_detect_nonce_mismatch(self):
        client = PyoidcFacade(
            ProviderConfiguration(provider_metadata=ProviderMetadata(issuer=self.ISSUER),
                                  client_metadata=ClientMetadata(client_id=self.CLIENT_ID)),
            redirect_uri='https://client.example.com/redirect')
        client.exchange_authorization_code = MagicMock(return_value=self.TOKEN_RESPONSE)
        auth_request = {'state': self.AUTH_RESPONSE['state'], 'nonce': 'other_nonce'}
        with pytest.raises(InvalidIdTokenError):
            AuthResponseHandler(client).process_auth_response(self.AUTH_RESPONSE, auth_request, {})

    def test_should_handle_auth_error_response(self, client_mock):
        with pytest.raises(AuthResponseErrorResponseError) as exc:
            AuthResponseHandler(client_mock).process_auth_response(AuthorizationErrorResponse(**self.ERROR_RESPONSE),
                                                                   self.AUTH_REQUEST,
                                                                   {})
        assert exc.value.error_response == self.ERROR_RESPONSE

    def test_should_handle_token_error_response(self, client_mock):
        client_mock.exchange_authorization_code.return_value = TokenErrorResponse(**self.ERROR_RESPONSE)
        with pytest.raises(AuthResponseErrorResponseError) as exc:
            AuthResponseHandler(client_mock).process_auth_response(AuthorizationResponse(**self.AUTH_RESPONSE),
                                                                   self.AUTH_REQUEST,
                                                                   {})
        assert exc.value.error_response == self.ERROR_RESPONSE

    def test_should_detect_mismatching_subject(self, client_mock):
        client_mock.exchange_authorization_code.return_value = AccessTokenResponse(**self.TOKEN_RESPONSE)
        client_mock.userinfo_request.return_value = OpenIDSchema(**{'sub': 'other_sub'})
        with pytest.raises(AuthResponseMismatchingSubjectError):
            AuthResponseHandler(client_mock).process_auth_response(AuthorizationResponse(**self.AUTH_RESPONSE),
                                                                   self.AUTH_REQUEST,
                                                                   {})

    def test_should_handle_auth_response_with_authorization_code(self, client_mock):
        client_mock.exchange_authorization_code.return_value = self.TOKEN_RESPONSE
        client_mock.userinfo_request.return_value = self.USERINFO_RESPONSE
        result = AuthResponseHandler(client_mock).process_auth_response(self.AUTH_RESPONSE,
                                                                        self.AUTH_REQUEST,
                                                                        {})
        assert result.access_token == 'test_token'
        assert result.expires_in == self.TOKEN_RESPONSE['expires_in']
        assert result.id_token_claims == self.TOKEN_RESPONSE['id_token'].to_dict()
        assert result.id_token_jwt == self.TOKEN_RESPONSE['id_token_jwt']
        assert result.userinfo_claims == self.USERINFO_RESPONSE.to_dict()
        assert result.refresh_token == self.TOKEN_RESPONSE['refresh_token']

    def test_should_handle_auth_response_without_authorization_code(self, client_mock):
        auth_response = AuthorizationResponse(**self.TOKEN_RESPONSE)
        auth_response['state'] = 'test_state'
        client_mock.userinfo_request.return_value = self.USERINFO_RESPONSE
        result = AuthResponseHandler(client_mock).process_auth_response(auth_response, self.AUTH_REQUEST, {})
        assert not client_mock.exchange_authorization_code.called
        assert result.access_token == 'test_token'
        assert result.expires_in == self.TOKEN_RESPONSE['expires_in']
        assert result.id_token_jwt == self.TOKEN_RESPONSE['id_token_jwt']
        assert result.id_token_claims == self.TOKEN_RESPONSE['id_token'].to_dict()
        assert result.userinfo_claims == self.USERINFO_RESPONSE.to_dict()
        assert result.refresh_token is None

    def test_should_handle_token_response_without_id_token(self, client_mock):
        token_response = {'access_token': 'test_token'}
        client_mock.exchange_authorization_code.return_value = AccessTokenResponse(**token_response)
        result = AuthResponseHandler(client_mock).process_auth_response(AuthorizationResponse(**self.AUTH_RESPONSE),
                                                                        self.AUTH_REQUEST,
                                                                        {})
        assert result.access_token == 'test_token'
        assert result.id_token_claims is None

    def test_should_handle_no_token_response(self, client_mock):
        client_mock.exchange_authorization_code.return_value = None
        client_mock.userinfo_request.return_value = None
        hybrid_auth_response = self.AUTH_RESPONSE.copy()
        hybrid_auth_response.update(self.TOKEN_RESPONSE)
        result = AuthResponseHandler(client_mock).process_auth_response(AuthorizationResponse(**hybrid_auth_response),
                                                                        self.AUTH_REQUEST,
                                                                        {})
        assert result.access_token == 'test_token'
        assert result.id_token_claims == self.TOKEN_RESPONSE['id_token'].to_dict()
        assert result.id_token_jwt == self.TOKEN_RESPONSE['id_token_jwt']

    @pytest.mark.parametrize('response_type, expected', [
        ('code', False),  # Authorization Code Flow
        ('id_token', True),  # Implicit Flow
        ('id_token token', True),  # Implicit Flow
        ('code id_token', True),  # Hybrid Flow
        ('code token', True),  # Hybrid Flow
        ('code id_token token', True)  # Hybrid Flow
    ])
    def test_expect_fragment_encoded_response_by_response_type(self, response_type, expected):
        assert AuthResponseHandler.expect_fragment_encoded_response({'response_type': response_type}) is expected

    @pytest.mark.parametrize('response_type, response_mode, expected', [
        ('code', 'fragment', True),
        ('id_token', 'query', False),
        ('code token', 'form_post', False),
    ])
    def test_expect_fragment_encoded_response_with_non_default_response_mode(self,
                                                                             response_type,
                                                                             response_mode,
                                                                             expected):
        auth_req = {'response_type': response_type, 'response_mode': response_mode}
        assert AuthResponseHandler.expect_fragment_encoded_response(auth_req) is expected



from oic.oauth2.message import AccessTokenResponse, CCAccessTokenRequest, MessageTuple, OauthMessageFactory


class CCMessageFactory(OauthMessageFactory):
    """Client Credential Request Factory."""
    token_endpoint = MessageTuple(CCAccessTokenRequest, AccessTokenResponse)



import base64

import pytest
import responses
from flask_pyoidc.provider_configuration import ProviderConfiguration, ClientRegistrationInfo, ProviderMetadata, \
    ClientMetadata, OIDCData
from oic.oic import Client
from oic.utils.authn.client import CLIENT_AUTHN_METHOD


class TestProviderConfiguration:
    PROVIDER_BASEURL = 'https://op.example.com'

    @staticmethod
    def provider_metadata(**kwargs):
        return ProviderMetadata(issuer='', authorization_endpoint='', jwks_uri='', **kwargs)

    def test_missing_provider_metadata_raises_exception(self):
        with pytest.raises(ValueError) as exc_info:
            ProviderConfiguration(client_registration_info=ClientRegistrationInfo())

        exc_message = str(exc_info.value)
        assert 'issuer' in exc_message
        assert 'provider_metadata' in exc_message

    def test_missing_client_metadata_raises_exception(self):
        with pytest.raises(ValueError) as exc_info:
            ProviderConfiguration(issuer=self.PROVIDER_BASEURL)

        exc_message = str(exc_info.value)
        assert 'client_registration_info' in exc_message
        assert 'client_metadata' in exc_message

    @responses.activate
    def test_should_fetch_provider_metadata_if_not_given(self):
        provider_metadata = {
            'issuer': self.PROVIDER_BASEURL,
            'authorization_endpoint': self.PROVIDER_BASEURL + '/auth',
            'jwks_uri': self.PROVIDER_BASEURL + '/jwks'
        }
        responses.add(responses.GET,
                      self.PROVIDER_BASEURL + '/.well-known/openid-configuration',
                      json=provider_metadata)

        provider_config = ProviderConfiguration(issuer=self.PROVIDER_BASEURL,
                                                client_registration_info=ClientRegistrationInfo())
        provider_config.ensure_provider_metadata(Client(CLIENT_AUTHN_METHOD))
        assert provider_config._provider_metadata['issuer'] == self.PROVIDER_BASEURL
        assert provider_config._provider_metadata['authorization_endpoint'] == self.PROVIDER_BASEURL + '/auth'
        assert provider_config._provider_metadata['jwks_uri'] == self.PROVIDER_BASEURL + '/jwks'

    def test_should_not_fetch_provider_metadata_if_given(self):
        provider_metadata = self.provider_metadata()
        provider_config = ProviderConfiguration(provider_metadata=provider_metadata,
                                                client_registration_info=ClientRegistrationInfo())

        provider_config.ensure_provider_metadata(Client(CLIENT_AUTHN_METHOD))
        assert provider_config._provider_metadata == provider_metadata

    @responses.activate
    def test_should_register_dynamic_client_if_client_registration_info_is_given(self):
        registration_endpoint = self.PROVIDER_BASEURL + '/register'
        redirect_uris = ['https://client.example.com/redirect',
                         'https://client.example.com/redirect2']
        post_logout_redirect_uris = ['https://client.example.com/logout']
        responses.add(responses.POST, registration_endpoint, json={
            'client_id': 'client1', 'client_secret': 'secret1',
            'redirect_uris': redirect_uris,
            'post_logout_redirect_uris': post_logout_redirect_uris})

        provider_config = ProviderConfiguration(
            provider_metadata=self.provider_metadata(registration_endpoint=registration_endpoint),
            client_registration_info=ClientRegistrationInfo(
                redirect_uris=redirect_uris,
                post_logout_redirect_uris=post_logout_redirect_uris))

        provider_config.register_client(Client(CLIENT_AUTHN_METHOD))
        assert provider_config._client_metadata['client_id'] == 'client1'
        assert provider_config._client_metadata['client_secret'] == 'secret1'
        assert provider_config._client_metadata['redirect_uris'] == redirect_uris
        assert provider_config._client_metadata[
                   'post_logout_redirect_uris'] == post_logout_redirect_uris

    def test_should_not_register_dynamic_client_if_client_metadata_is_given(self):
        client_metadata = ClientMetadata(client_id='client1',
                                         client_secret='secret1')
        provider_config = ProviderConfiguration(provider_metadata=self.provider_metadata(),
                                                client_metadata=client_metadata)
        provider_config.register_client(None)
        assert provider_config._client_metadata == client_metadata

    def test_should_raise_exception_for_non_registered_client_when_missing_registration_endpoint(self):
        provider_config = ProviderConfiguration(provider_metadata=self.provider_metadata(),
                                                client_registration_info=ClientRegistrationInfo())
        assert provider_config._provider_metadata['registration_endpoint'] is None
        with pytest.raises(ValueError):
            provider_config.register_client(None)

    @responses.activate
    def test_register_client_should_register_dynamic_client_if_initial_access_token_present(self):

        registration_endpoint = self.PROVIDER_BASEURL + '/register'
        redirect_uris = ['https://client.example.com/redirect',
                         'https://client.example.com/redirect2']
        post_logout_redirect_uris = ['https://client.example.com/logout']
        client_registration_response = {
            'client_id': 'client1',
            'client_secret': 'secret1',
            'client_name': 'Test Client',
            'redirect_uris': redirect_uris,
            'post_logout_redirect_uris': post_logout_redirect_uris,
            'registration_client_uri': 'https://op.example.com/register/client1',
            'registration_access_token': 'registration_access_token1'
        }
        responses.add(responses.POST, registration_endpoint, json=client_registration_response)
        provider_config = ProviderConfiguration(
            provider_metadata=self.provider_metadata(registration_endpoint=registration_endpoint),
            client_registration_info=ClientRegistrationInfo(
                client_name='Test Client',
                redirect_uris=redirect_uris,
                post_logout_redirect_uris=post_logout_redirect_uris,
                registration_token='initial_access_token'))

        provider_config.register_client(Client(CLIENT_AUTHN_METHOD))

        assert provider_config._client_metadata['client_id'] == 'client1'
        assert provider_config._client_metadata['client_secret'] == 'secret1'
        assert provider_config._client_metadata['client_name'] == 'Test Client'
        assert provider_config._client_metadata['registration_client_uri'] == 'https://op.example.com/register/client1'
        assert provider_config._client_metadata['registration_access_token'] == 'registration_access_token1'
        assert provider_config._client_metadata['redirect_uris'] == redirect_uris
        assert provider_config._client_metadata[
                   'post_logout_redirect_uris'] == post_logout_redirect_uris
        assert responses.calls[0].request.headers['Authorization'] == \
               f"Bearer {base64.b64encode('initial_access_token'.encode()).decode()}"

    @responses.activate
    def test_register_client_should_register_client_even_if_post_logout_redirect_uris_missing(self):
        registration_endpoint = self.PROVIDER_BASEURL + '/register'
        redirect_uris = ['https://client.example.com/redirect',
                         'https://client.example.com/redirect2']
        client_registration_response = {
            'client_id': 'client1',
            'client_secret': 'secret1',
            'client_name': 'Test Client',
            'redirect_uris': redirect_uris,
            'registration_client_uri': 'https://op.example.com/register/client1',
            'registration_access_token': 'registration_access_token1'
        }
        responses.add(responses.POST, registration_endpoint, json=client_registration_response)
        provider_config = ProviderConfiguration(
            provider_metadata=self.provider_metadata(registration_endpoint=registration_endpoint),
            client_registration_info=ClientRegistrationInfo(
                client_name='Test Client',
                redirect_uris=redirect_uris,
                post_logout_redirect_uris=[],
                registration_token='initial_access_token'))

        provider_config.register_client(client=Client(CLIENT_AUTHN_METHOD))

        assert provider_config._client_metadata['client_id'] == 'client1'
        assert provider_config._client_metadata['client_secret'] == 'secret1'
        assert provider_config._client_metadata['client_name'] == 'Test Client'
        assert provider_config._client_metadata['registration_client_uri'] == 'https://op.example.com/register/client1'
        assert provider_config._client_metadata['registration_access_token'] == 'registration_access_token1'
        assert provider_config._client_metadata['redirect_uris'] == redirect_uris
        assert provider_config._client_metadata.get('post_logout_redirect_uris') is None


class TestOIDCData:
    def test_client_secret_should_not_be_in_string_representation(self):
        client_secret = 'secret123456'
        client_metadata = OIDCData(client_id='client1', client_secret=client_secret)
        assert client_secret not in str(client_metadata)
        assert client_secret in repr(client_metadata)

    def test_copy_should_overwrite_existing_value(self):
        data = OIDCData(abc='xyz')
        copy_data = data.copy(qwe='rty', abc='123')
        assert copy_data == {'abc': '123', 'qwe': 'rty'}

    def test_del_and_len(self):
        data = OIDCData(abc='xyz', qwe='rty')
        assert len(data) == 2
        del data['qwe']
        assert data.to_dict() == {'abc': 'xyz'}



import collections
import logging

from oic.exception import PyoidcError

logger = logging.getLogger(__name__)

AuthenticationResult = collections.namedtuple('AuthenticationResult',
                                              [
                                                  'access_token',
                                                  'expires_in',
                                                  'id_token_claims',
                                                  'id_token_jwt',
                                                  'userinfo_claims',
                                                  'refresh_token'
                                              ])


class AuthResponseProcessError(ValueError):
    pass


class AuthResponseUnexpectedStateError(AuthResponseProcessError):
    pass


class InvalidIdTokenError(AuthResponseProcessError):
    pass


class AuthResponseMismatchingSubjectError(AuthResponseProcessError):
    pass


class AuthResponseErrorResponseError(AuthResponseProcessError):
    def __init__(self, error_response):
        """
        Args:
            error_response (Mapping[str, str]): OAuth error response containing 'error' and 'error_description'
        """
        self.error_response = error_response


class AuthResponseHandler:
    def __init__(self, client):
        """
        Args:
            client (flask_pyoidc.pyoidc_facade.PyoidcFacade): Client proxy to make requests to the provider
        """
        self._client = client

    def process_auth_response(self, auth_response, auth_request, extra_token_args):
        """
        Args:
            auth_response (Union[AuthorizationResponse, AuthorizationErrorResponse]): parsed OIDC auth response
            auth_request (Mapping[str, str]): original OIDC auth request
            extra_token_args  (Mapping[str, Any]): extra arguments to pass to pyoidc
        Returns:
            AuthenticationResult: All relevant data associated with the authenticated user
        """
        if 'error' in auth_response:
            raise AuthResponseErrorResponseError(auth_response.to_dict())

        if auth_response['state'] != auth_request['state']:
            raise AuthResponseUnexpectedStateError()

        # implicit/hybrid flow may return tokens in the auth response
        access_token = auth_response.get('access_token', None)
        expires_in = auth_response.get('expires_in', None)
        id_token_claims = auth_response['id_token'].to_dict() if 'id_token' in auth_response else None
        id_token_jwt = auth_response.get('id_token_jwt', None)
        refresh_token = None  # but never refresh token

        if 'code' in auth_response:
            token_resp = self._client.exchange_authorization_code(auth_response['code'],
                                                                  auth_response['state'],
                                                                  extra_token_args)
            if token_resp:
                if 'error' in token_resp:
                    raise AuthResponseErrorResponseError(token_resp.to_dict())

                access_token = token_resp['access_token']
                expires_in = token_resp.get('expires_in', None)
                refresh_token = token_resp.get('refresh_token', None)

                if 'id_token' in token_resp:
                    id_token = token_resp['id_token']
                    logger.debug('received id token: %s', id_token.to_json())

                    try:
                        self._client.verify_id_token(id_token, auth_request)
                    except PyoidcError as ex:
                        raise InvalidIdTokenError(str(ex)) from ex

                    id_token_claims = id_token.to_dict()
                    id_token_jwt = token_resp.get('id_token_jwt')

        # do userinfo request
        userinfo = self._client.userinfo_request(access_token)
        userinfo_claims = None
        if userinfo:
            userinfo_claims = userinfo.to_dict()

        if id_token_claims and userinfo_claims and userinfo_claims['sub'] != id_token_claims['sub']:
            raise AuthResponseMismatchingSubjectError('The \'sub\' of userinfo does not match \'sub\' of ID Token.')

        return AuthenticationResult(access_token,
                                    expires_in,
                                    id_token_claims,
                                    id_token_jwt,
                                    userinfo_claims,
                                    refresh_token)

    @classmethod
    def expect_fragment_encoded_response(cls, auth_request):
        if 'response_mode' in auth_request:
            return auth_request['response_mode'] == 'fragment'

        response_type = set(auth_request['response_type'].split(' '))
        is_implicit_flow = response_type == {'id_token'} or \
                           response_type == {'id_token', 'token'}
        is_hybrid_flow = response_type == {'code', 'id_token'} or \
                         response_type == {'code', 'token'} or \
                         response_type == {'code', 'id_token', 'token'}

        return is_implicit_flow or is_hybrid_flow



#   Copyright 2020 Samuel Gulliksson
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0

#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import warnings
from urllib.parse import urlparse


class RedirectUriConfig:
    def __init__(self, full_uri, endpoint):
        self.full_uri = full_uri
        self.endpoint = endpoint

    def __eq__(self, other):
        return self.full_uri == other.full_uri and self.endpoint == other.endpoint

    def __str__(self):
        return '(' + self.full_uri + ', ' + self.endpoint + ')'

    def __repr__(self):
        return str(self)

    @classmethod
    def from_config(cls, config):
        if 'OIDC_REDIRECT_URI' in config:
            return cls(*RedirectUriConfig._parse_redirect_uri(config['OIDC_REDIRECT_URI']))

        return cls(*RedirectUriConfig._parse_legacy_config(config))

    @staticmethod
    def _parse_redirect_uri(redirect_uri):
        parsed = urlparse(redirect_uri)
        endpoint = parsed.path.lstrip('/')
        return redirect_uri, endpoint

    @staticmethod
    def _parse_legacy_config(config):
        redirect_domain = config.get('OIDC_REDIRECT_DOMAIN', config.get('SERVER_NAME'))
        if not redirect_domain:
            raise ValueError("'OIDC_REDIRECT_URI' must be configured.")

        scheme = config.get('PREFERRED_URL_SCHEME', 'http')

        warnings.warn(
            "Please use 'OIDC_REDIRECT_URI' to configure the redirect_uri for flask-pyoidc. 'OIDC_REDIRECT_DOMAIN' and 'OIDC_REDIRECT_ENDPOINT' have been deprecated.",
            DeprecationWarning,
            stacklevel=2
        )

        endpoint = config.get('OIDC_REDIRECT_ENDPOINT', 'redirect_uri').lstrip('/')
        full_uri = scheme + '://' + redirect_domain + '/' + endpoint

        return full_uri, endpoint



import time
from unittest.mock import patch

import pytest
from flask_pyoidc.user_session import UserSession, UninitialisedSession


class TestUserSession:
    PROVIDER_NAME = 'test_provider'

    def initialised_session(self, session_storage):
        return UserSession(session_storage, self.PROVIDER_NAME)

    def test_initialising_session_with_existing_user_session_should_preserve_state(self):
        storage = {}
        session1 = UserSession(storage, self.PROVIDER_NAME)
        session1.update()
        assert session1.is_authenticated() is True
        assert session1.current_provider == self.PROVIDER_NAME

        session2 = UserSession(storage, self.PROVIDER_NAME)
        assert session2.is_authenticated() is True
        assert session2.current_provider == self.PROVIDER_NAME

        session3 = UserSession(storage)
        assert session3.is_authenticated() is True
        assert session3.current_provider == self.PROVIDER_NAME

    def test_initialising_session_with_new_provider_name_should_reset_session(self):
        storage = {}
        session1 = UserSession(storage, 'provider1')
        session1.update()
        assert session1.is_authenticated() is True
        session2 = UserSession(storage, 'provider2')
        assert session2.is_authenticated() is False

    def test_unauthenticated_session(self):
        assert self.initialised_session({}).is_authenticated() is False

    def test_authenticated_session(self):
        assert self.initialised_session({'last_authenticated': 1234}).is_authenticated() is True

    def test_should_not_refresh_if_not_supported(self):
        assert self.initialised_session({}).should_refresh() is False

    def test_should_not_refresh_if_authenticated_within_refresh_interval(self):
        refresh_interval = 10
        session = self.initialised_session({'last_session_refresh': time.time() + (refresh_interval - 1)})
        assert session.should_refresh(refresh_interval) is False

    def test_should_refresh_if_supported_and_necessary(self):
        refresh_interval = 10
        # authenticated too far in the past
        session_storage = {'last_session_refresh': time.time() - (refresh_interval + 1)}
        assert self.initialised_session(session_storage).should_refresh(refresh_interval) is True

    def test_should_not_refresh_if_not_previously_authenticated(self):
        assert self.initialised_session({}).should_refresh(10) is False

    @pytest.mark.parametrize('data', [
        {'access_token': 'test_access_token'},
        {'id_token': {'iss': 'issuer1', 'sub': 'user1', 'aud': 'client1', 'exp': 1235, 'iat': 1234}},
        {'id_token_jwt': 'eyJh.eyJz.SflK'},
        {'userinfo': {'sub': 'user1', 'name': 'Test User'}},
    ])
    @patch('time.time')
    def test_update(self, time_mock, data):
        storage = {}
        auth_time = 1234
        time_mock.return_value = auth_time

        self.initialised_session(storage).update(**data)

        expected_session_data = {
            'last_authenticated': auth_time,
            'last_session_refresh': auth_time,
            'current_provider': self.PROVIDER_NAME
        }
        expected_session_data.update(**data)
        assert storage == expected_session_data

    def test_update_should_use_auth_time_from_id_token_if_it_exists(self):
        auth_time = 1234
        session = self.initialised_session({})
        session.update(id_token={'auth_time': auth_time})
        assert session.last_authenticated == auth_time

    @patch('time.time')
    def test_update_should_update_last_session_refresh_timestamp(self, time_mock):
        now_timestamp = 1234
        time_mock.return_value = now_timestamp
        data = {}
        session = self.initialised_session(data)
        session.update()
        assert data['last_session_refresh'] == now_timestamp

    def test_trying_to_update_uninitialised_session_should_throw_exception(self):
        with pytest.raises(UninitialisedSession):
            UserSession(session_storage={}).update()

    def test_clear(self):
        expected_data = {'initial data': 'should remain'}
        session_storage = expected_data.copy()

        session = self.initialised_session(session_storage)
        session.update(access_token='access_token', expires_in=3600, id_token={'sub': 'user1'}, id_token_jwt='eyJh.eyJz.SflK', userinfo={'sub': 'user1}'}, refresh_token='refresh_token')
        session.clear()

        assert session_storage == expected_data

    def test_access_token_expiry(self):
        session = self.initialised_session({})
        expires_in = 3600
        session.update(expires_in=expires_in)
        assert session.access_token_expires_at == int(time.time()) + expires_in



import logging
from typing import Any, Mapping

from oic.extension.client import Client as ClientExtension
from oic.extension.message import TokenIntrospectionResponse
from oic.oauth2 import Client as Oauth2Client
from oic.oauth2.message import AccessTokenResponse
from oic.oic import Client
from oic.oic import Token
from oic.oic.message import AuthorizationResponse, ProviderConfigurationResponse, RegistrationResponse
from oic.utils.authn.client import CLIENT_AUTHN_METHOD

from .message_factory import CCMessageFactory

logger = logging.getLogger(__name__)


class PyoidcFacade:
    """
    Wrapper around pyoidc library, coupled with config for a simplified API for flask-pyoidc.
    """

    def __init__(self, provider_configuration, redirect_uri):
        """
        Args:
            provider_configuration (flask_pyoidc.provider_configuration.ProviderConfiguration)
        """
        self._provider_configuration = provider_configuration
        self._client = Client(client_authn_method=CLIENT_AUTHN_METHOD,
                              settings=provider_configuration.client_settings)
        # Token Introspection is implemented under extension sub-package of
        # the client in pyoidc.
        self._client_extension = ClientExtension(client_authn_method=CLIENT_AUTHN_METHOD,
                                                 settings=provider_configuration.client_settings)
        # Client Credentials Flow is implemented under oauth2 sub-package of
        # the client in pyoidc.
        self._oauth2_client = Oauth2Client(client_authn_method=CLIENT_AUTHN_METHOD,
                                           message_factory=CCMessageFactory,
                                           settings=self._provider_configuration.client_settings)

        provider_metadata = provider_configuration.ensure_provider_metadata(self._client)
        self._client.handle_provider_config(ProviderConfigurationResponse(**provider_metadata.to_dict()),
                                            provider_metadata['issuer'])

        if self._provider_configuration.registered_client_metadata:
            client_metadata = self._provider_configuration.registered_client_metadata.to_dict()
            client_metadata.update(redirect_uris=[redirect_uri])
            self._store_registration_info(client_metadata)

        self._redirect_uri = redirect_uri

    def _store_registration_info(self, client_metadata):
        registration_response = RegistrationResponse(**client_metadata)
        self._client.store_registration_info(registration_response)
        self._client_extension.store_registration_info(registration_response)
        # Set client_id and client_secret for _oauth2_client. This is used
        # by Client Credentials Flow.
        self._oauth2_client.client_id = registration_response['client_id']
        self._oauth2_client.client_secret = registration_response['client_secret']

    def is_registered(self):
        return bool(self._provider_configuration.registered_client_metadata)

    def register(self):
        client_metadata = self._provider_configuration.register_client(self._client)
        logger.debug(f'client registration response: {client_metadata}')
        self._store_registration_info(client_metadata)

    def authentication_request(self, state, nonce, extra_auth_params):
        """

        Args:
            state (str): authentication request parameter 'state'
            nonce (str): authentication request parameter 'nonce'
            extra_auth_params (Mapping[str, str]): extra authentication request parameters
        Returns:
            AuthorizationRequest: the authentication request
        """
        args = {
            'client_id': self._client.client_id,
            'response_type': 'code',
            'scope': ['openid'],
            'redirect_uri': self._redirect_uri,
            'state': state,
            'nonce': nonce,
        }

        args.update(self._provider_configuration.auth_request_params)
        args.update(extra_auth_params)
        auth_request = self._client.construct_AuthorizationRequest(request_args=args)
        logger.debug('sending authentication request: %s', auth_request.to_json())

        return auth_request

    def login_url(self, auth_request):
        """
        Args:
            auth_request (AuthorizationRequest): authentication request
        Returns:
            str: Authentication request as a URL to redirect the user to the provider.
        """
        return auth_request.request(self._client.authorization_endpoint)

    def parse_authentication_response(self, response_params):
        """
        Parameters
        ----------
        response_params: Mapping[str, str]
            authentication response parameters.

        Returns
        -------
        Union[AuthorizationResponse, AuthorizationErrorResponse]
            The parsed authorization response.
        """
        auth_resp = self._client.parse_response(AuthorizationResponse, info=response_params, sformat='dict')
        if 'id_token' in response_params:
            auth_resp['id_token_jwt'] = response_params['id_token']
        return auth_resp

    def exchange_authorization_code(self, authorization_code: str, state: str, extra_token_args: Mapping[str, Any]):
        """Requests tokens from an authorization code.

        Parameters
        ----------
        authorization_code: str
            authorization code issued to client after user authorization
        state: str
            state is used to keep track of responses to outstanding requests.
        extra_token_args: Mapping[str, Any]
            extra arguments to pass to pyoidc

        Returns
        -------
        Union[AccessTokenResponse, TokenErrorResponse, None]
            The parsed token response, or None if no token request was performed.
        """
        if not self._client.token_endpoint:
            return None

        request_args = {
            'grant_type': 'authorization_code',
            'code': authorization_code,
            'redirect_uri': self._redirect_uri
        }
        logger.debug('making token request: %s', request_args)
        client_auth_method = self._client.registration_response.get('token_endpoint_auth_method',
                                                                    'client_secret_basic')
        token_response = self._client.do_access_token_request(state=state,
                                                              request_args=request_args,
                                                              authn_method=client_auth_method,
                                                              endpoint=self._client.token_endpoint,
                                                              **extra_token_args)
        logger.info('Received token response.')

        return token_response

    def verify_id_token(self, id_token, auth_request):
        """
        Verifies the ID Token.

        Args:
            id_token (Mapping[str, str]): ID token claims
            auth_request (Mapping[str, str]): original authentication request parameters to validate against
                (nonce, acr_values, max_age, etc.)

        Raises:
            PyoidcError: If the ID token is invalid.

        """
        self._client.verify_id_token(id_token, auth_request)

    def refresh_token(self, refresh_token: str):
        """Requests new tokens using a refresh token.

        Parameters
        ----------
        refresh_token: str
            refresh token issued to client after user authorization.

        Returns
        -------
        Union[AccessTokenResponse, TokenErrorResponse, None]
            The parsed token response, or None if no token request was performed.
        """
        request_args = {
            'grant_type': 'refresh_token',
            'refresh_token': refresh_token,
            'redirect_uri': self._redirect_uri
        }
        client_auth_method = self._client.registration_response.get('token_endpoint_auth_method',
                                                                    'client_secret_basic')
        return self._client.do_access_token_refresh(request_args=request_args,
                                                    authn_method=client_auth_method,
                                                    token=Token(resp={'refresh_token': refresh_token}),
                                                    endpoint=self._client.token_endpoint
                                                    )

    def userinfo_request(self, access_token: str):
        """Retrieves ID token.

        Parameters
        ----------
        access_token: str
            Bearer access token to use when fetching userinfo.

        Returns
        -------
        Union[OpenIDSchema, UserInfoErrorResponse, ErrorResponse, None]
        """
        http_method = self._provider_configuration.userinfo_endpoint_method
        if not access_token or http_method is None or not self._client.userinfo_endpoint:
            return None

        logger.debug('making userinfo request')
        userinfo_response = self._client.do_user_info_request(method=http_method, token=access_token)
        logger.debug('received userinfo response: %s', userinfo_response)

        return userinfo_response

    def _token_introspection_request(self, access_token: str) -> TokenIntrospectionResponse:
        """Make token introspection request.

        Parameters
        ----------
        access_token: str
            Access token to be validated.

        Returns
        -------
        TokenIntrospectionResponse
            Response object contains result of the token introspection.
        """
        request_args = {
            'token': access_token,
            'token_type_hint': 'access_token'
        }
        client_auth_method = self._client.registration_response.get('introspection_endpoint_auth_method',
                                                                    'client_secret_basic')
        logger.info('making token introspection request')
        token_introspection_response = self._client_extension.do_token_introspection(
            request_args=request_args, authn_method=client_auth_method, endpoint=self._client.introspection_endpoint)

        return token_introspection_response

    def client_credentials_grant(self, scope: list = None, **kwargs) -> AccessTokenResponse:
        """Public method to request access_token using client_credentials flow.
        This is useful for service to service communication where user-agent is
        not available which is required in authorization code flow. Your
        service can request access_token in order to access APIs of other
        services.

        On API call, token introspection will ensure that only valid token can
        be used to access your APIs.

        Parameters
        ----------
        scope: list, optional
            List of scopes to be requested.
        **kwargs : dict, optional
            Extra arguments to client credentials flow.

        Returns
        -------
        AccessTokenResponse

        Examples
        --------
        ::

            auth = OIDCAuthentication({'default': provider_config},
                                      access_token_required=True)
            auth.init_app(app)
            auth.clients['default'].client_credentials_grant()

        Optionally, you can specify scopes for the access token.

        ::

            auth.clients['default'].client_credentials_grant(
                scope=['read', 'write'])

        You can also specify extra keyword arguments to client credentials flow.

        ::

            auth.clients['default'].client_credentials_grant(
                scope=['read', 'write'], audience=['client_id1', 'client_id2'])
        """
        request_args = {
            'grant_type': 'client_credentials',
            **kwargs
        }
        if scope:
            request_args['scope'] = ' '.join(scope)
        client_auth_method = self._client.registration_response.get('token_endpoint_auth_method',
                                                                    'client_secret_basic')
        access_token = self._oauth2_client.do_access_token_request(request_args=request_args,
                                                                   authn_method=client_auth_method,
                                                                   endpoint=self._client.token_endpoint)
        return access_token

    @property
    def session_refresh_interval_seconds(self):
        return self._provider_configuration.session_refresh_interval_seconds

    @property
    def provider_end_session_endpoint(self):
        provider_metadata = self._provider_configuration.ensure_provider_metadata(self._client)
        return provider_metadata.get('end_session_endpoint')

    @property
    def post_logout_redirect_uris(self):
        return self._client.registration_response.get('post_logout_redirect_uris')



import collections.abc
import logging

import requests
from oic.oic import Client
from oic.utils.settings import ClientSettings

logger = logging.getLogger(__name__)


class OIDCData(collections.abc.MutableMapping):
    """
    Basic OIDC data representation providing validation of required fields.
    """

    def __init__(self, *args, **kwargs):
        """
        Args:
            args (List[Tuple[String, String]]): key-value pairs to store
            kwargs (Dict[string, string]): key-value pairs to store
        """
        self.store = {}
        self.update(dict(*args, **kwargs))

    def __getitem__(self, key):
        return self.store[key]

    def __setitem__(self, key, value):
        self.store[key] = value

    def __delitem__(self, key):
        del self.store[key]

    def __iter__(self):
        return iter(self.store)

    def __len__(self):
        return len(self.store)

    def __str__(self):
        data = self.store.copy()
        if 'client_secret' in data:
            data['client_secret'] = '<masked>'
        return str(data)

    def __repr__(self):
        return str(self.store)

    def __bool__(self):
        return True

    def copy(self, **kwargs):
        values = self.to_dict()
        values.update(kwargs)
        return self.__class__(**values)

    def to_dict(self):
        return self.store.copy()


class ProviderMetadata(OIDCData):

    def __init__(self,
                 issuer=None,
                 authorization_endpoint=None,
                 jwks_uri=None,
                 token_endpoint=None,
                 userinfo_endpoint=None,
                 introspection_endpoint=None,
                 registration_endpoint=None,
                 **kwargs):
        """OpenID Providers have metadata describing their configuration.

        Parameters
        ----------
        issuer: str, Optional
            OP Issuer Identifier.
        authorization_endpoint: str, Optional
            URL of the OP's OAuth 2.0 Authorization Endpoint.
        jwks_uri: str, Optional
            URL of the OP's JSON Web Key Set [JWK] document.
        token_endpoint: str, Optional
            URL of the OP's OAuth 2.0 Token Endpoint.
        userinfo_endpoint: str, Optional
            URL of the OP's UserInfo Endpoint.
        introspection_endpoint: str, Optional
            URL of the OP's token introspection endpoint.
        registration_endpoint: str, Optional
            URL of the OP's Dynamic Client Registration Endpoint.
        **kwargs : dict, Optional
            Extra arguments to [OpenID Provider Metadata](https://openid.net/specs/openid-connect-discovery-1_0.html#ProviderMetadata)
        """
        super().__init__(issuer=issuer, authorization_endpoint=authorization_endpoint,
                         token_endpoint=token_endpoint, userinfo_endpoint=userinfo_endpoint,
                         jwks_uri=jwks_uri, introspection_endpoint=introspection_endpoint,
                         registration_endpoint=registration_endpoint, **kwargs)


class ClientRegistrationInfo(OIDCData):
    pass


class ClientMetadata(OIDCData):
    def __init__(self, client_id=None, client_secret=None, **kwargs):
        """
        Args:
            client_id (str): client identifier representing the client
            client_secret (str): client secret to authenticate the client with
                the OP
            kwargs (dict): key-value pairs
        """
        super().__init__(client_id=client_id, client_secret=client_secret, **kwargs)


class ProviderConfiguration:
    """
    Metadata for communicating with an OpenID Connect Provider (OP).

    Attributes:
        auth_request_params (dict): Extra parameters, as key-value pairs, to include in the query parameters
            of the authentication request
        registered_client_metadata (ClientMetadata): The client metadata registered with the provider.
        requests_session (requests.Session): Requests object to use when communicating with the provider.
        session_refresh_interval_seconds (int): Number of seconds between updates of user data (tokens, user data, etc.)
            fetched from the provider. If `None` is specified, no silent updates should be made user data will be made.
        userinfo_endpoint_method (str): HTTP method ("GET" or "POST") to use when making the UserInfo Request. If
            `None` is specified, no UserInfo Request will be made.
    """

    DEFAULT_REQUEST_TIMEOUT = 5

    def __init__(self,
                 issuer=None,
                 provider_metadata=None,
                 userinfo_http_method='GET',
                 client_registration_info=None,
                 client_metadata=None,
                 auth_request_params=None,
                 session_refresh_interval_seconds=None,
                 requests_session=None):
        """
        Args:
            issuer (str): OP Issuer Identifier. If this is specified discovery will be used to fetch the provider
                metadata, otherwise `provider_metadata` must be specified.
            provider_metadata (ProviderMetadata): OP metadata,
            userinfo_http_method (Optional[str]): HTTP method (GET or POST) to use when sending the UserInfo Request.
                If `none` is specified, no userinfo request will be sent.
            client_registration_info (ClientRegistrationInfo): Client metadata to register your app
                dynamically with the provider. Either this or `registered_client_metadata` must be specified.
            client_metadata (ClientMetadata): Client metadata if your app is statically
                registered with the provider. Either this or `client_registration_info` must be specified.
            auth_request_params (dict): Extra parameters that should be included in the authentication request.
            session_refresh_interval_seconds (int): Length of interval (in seconds) between attempted user data
                refreshes.
            requests_session (requests.Session): custom requests object to allow for example retry handling, etc.
        """

        if not issuer and not provider_metadata:
            raise ValueError("Specify either 'issuer' or 'provider_metadata'.")

        if not client_registration_info and not client_metadata:
            raise ValueError("Specify either 'client_registration_info' or 'client_metadata'.")

        self._issuer = issuer
        self._provider_metadata = provider_metadata

        self._client_registration_info = client_registration_info
        self._client_metadata = client_metadata

        self.userinfo_endpoint_method = userinfo_http_method
        self.auth_request_params = auth_request_params or {}
        self.session_refresh_interval_seconds = session_refresh_interval_seconds
        # For session persistence
        self.client_settings = ClientSettings(timeout=self.DEFAULT_REQUEST_TIMEOUT,
                                              requests_session=requests_session or requests.Session())

    def ensure_provider_metadata(self, client: Client):
        if not self._provider_metadata:
            resp = client.provider_config(self._issuer)
            logger.debug(f'Received discovery response: {resp.to_dict()}')

            self._provider_metadata = ProviderMetadata(**resp.to_dict())

        return self._provider_metadata

    @property
    def registered_client_metadata(self):
        return self._client_metadata

    def register_client(self, client: Client):

        if not self._client_metadata:
            if not self._provider_metadata['registration_endpoint']:
                raise ValueError("Can't use dynamic client registration, provider metadata is missing "
                                 "'registration_endpoint'.")

            registration_request = self._client_registration_info.to_dict()

            # Send request to register the client dynamically.
            registration_response = client.register(
                url=self._provider_metadata['registration_endpoint'],
                **registration_request)
            logger.info('Received registration response.')
            self._client_metadata = ClientMetadata(
                **registration_response.to_dict())

        return self._client_metadata



import json
import logging
import time
from datetime import datetime
from http.cookies import SimpleCookie
from unittest.mock import MagicMock, patch
from urllib.parse import parse_qsl, urlparse, urlencode

import flask
import pytest
import responses
from flask import Flask
from flask_pyoidc import OIDCAuthentication
from flask_pyoidc.provider_configuration import (ProviderConfiguration, ProviderMetadata, ClientMetadata,
                                                 ClientRegistrationInfo)
from flask_pyoidc.redirect_uri_config import RedirectUriConfig
from flask_pyoidc.user_session import UserSession
from jwkest import jws
from oic.oic import AuthorizationResponse
from oic.oic.message import IdToken
from werkzeug.exceptions import Forbidden, Unauthorized
from werkzeug.routing import BuildError

from .util import signed_id_token


class TestOIDCAuthentication:
    PROVIDER_BASEURL = 'https://op.example.com'
    PROVIDER_NAME = 'test_provider'
    CLIENT_ID = 'client1'
    CLIENT_DOMAIN = 'client.example.com'
    CALLBACK_RETURN_VALUE = 'callback called successfully'

    @pytest.fixture(autouse=True)
    def create_flask_app(self):
        self.app = Flask(__name__)
        self.app.config.update({'SERVER_NAME': self.CLIENT_DOMAIN, 'SECRET_KEY': 'test_key', 'OIDC_CLOCK_SKEW': 10})

    def init_app(self, provider_metadata_extras=None, client_metadata_extras=None, **kwargs):
        required_provider_metadata = {
            'issuer': self.PROVIDER_BASEURL,
            'authorization_endpoint': self.PROVIDER_BASEURL + '/auth',
            'jwks_uri': self.PROVIDER_BASEURL + '/jwks'
        }
        if provider_metadata_extras:
            required_provider_metadata.update(provider_metadata_extras)
        provider_metadata = ProviderMetadata(**required_provider_metadata)

        required_client_metadata = {
            'client_id': self.CLIENT_ID,
            'client_secret': 'secret1'
        }
        if client_metadata_extras:
            required_client_metadata.update(client_metadata_extras)
        client_metadata = ClientMetadata(**required_client_metadata)

        provider_configurations = {self.PROVIDER_NAME: ProviderConfiguration(provider_metadata=provider_metadata,
                                                                             client_metadata=client_metadata,
                                                                             **kwargs)}
        authn = OIDCAuthentication(provider_configurations)
        authn.init_app(self.app)
        return authn

    def get_view_mock(self, name='test_callback'):
        mock = MagicMock()
        mock.__name__ = name
        mock.return_value = self.CALLBACK_RETURN_VALUE
        return mock

    def assert_auth_redirect(self, auth_redirect):
        assert auth_redirect.status_code == 302
        assert auth_redirect.location.startswith(self.PROVIDER_BASEURL)

    def assert_view_mock(self, callback_mock, result):
        assert callback_mock.called
        assert result == self.CALLBACK_RETURN_VALUE

    def test_explicit_redirect_uri_config_should_be_preferred(self):
        redirect_uri_config = RedirectUriConfig('https://example.com/redirect_uri', 'redirect_uri')
        assert OIDCAuthentication({}, self.app, redirect_uri_config)._redirect_uri_config == redirect_uri_config

    def test_explicit_redirect_uri_config_should_be_preserved_after_init_app(self):
        redirect_uri_config = RedirectUriConfig('https://example.com/redirect_uri', 'redirect_uri')
        authn = OIDCAuthentication({}, None, redirect_uri_config)
        assert authn._redirect_uri_config == redirect_uri_config
        authn.init_app(self.app)
        assert authn._redirect_uri_config == redirect_uri_config

    def test_should_authenticate_if_no_session(self):
        authn = self.init_app()
        view_mock = self.get_view_mock()
        with self.app.test_request_context('/'):
            auth_redirect = authn.oidc_auth(self.PROVIDER_NAME)(view_mock)()

        self.assert_auth_redirect(auth_redirect)
        assert not view_mock.called

    def test_should_not_authenticate_if_session_exists(self):
        authn = self.init_app()
        view_mock = self.get_view_mock()
        with self.app.test_request_context('/'):
            UserSession(flask.session, self.PROVIDER_NAME).update()
            result = authn.oidc_auth(self.PROVIDER_NAME)(view_mock)()
        self.assert_view_mock(view_mock, result)

    def test_reauthenticate_silent_if_session_expired(self):
        authn = self.init_app(session_refresh_interval_seconds=1)
        view_mock = self.get_view_mock()
        with self.app.test_request_context('/'):
            now = time.time()
            with patch('time.time') as time_mock:
                time_mock.return_value = now - 1  # authenticated in the past
                UserSession(flask.session, self.PROVIDER_NAME).update()
            auth_redirect = authn.oidc_auth(self.PROVIDER_NAME)(view_mock)()

        self.assert_auth_redirect(auth_redirect)
        assert 'prompt=none' in auth_redirect.location  # ensure silent auth is used
        assert not view_mock.called

    def test_dont_reauthenticate_silent_if_session_not_expired(self):
        authn = self.init_app(session_refresh_interval_seconds=999)
        view_mock = self.get_view_mock()
        with self.app.test_request_context('/'):
            UserSession(flask.session, self.PROVIDER_NAME).update()  # freshly authenticated
            result = authn.oidc_auth(self.PROVIDER_NAME)(view_mock)()
        self.assert_view_mock(view_mock, result)

    @pytest.mark.parametrize('response_type,expected', [
        ('code', False),
        ('id_token token', True)
    ])
    def test_expected_auth_response_mode_is_set(self, response_type, expected):
        authn = self.init_app(auth_request_params={'response_type': response_type})
        view_mock = self.get_view_mock()
        with self.app.test_request_context('/'):
            auth_redirect = authn.oidc_auth(self.PROVIDER_NAME)(view_mock)()
            assert flask.session['fragment_encoded_response'] is expected
        self.assert_auth_redirect(auth_redirect)

    @responses.activate
    @pytest.mark.parametrize('post_logout_redirect_uris', [
        None,
        ['https://example.com/post_logout']
    ])
    def test_should_register_client_if_not_registered_before(self, post_logout_redirect_uris):
        registration_endpoint = self.PROVIDER_BASEURL + '/register'
        provider_metadata = ProviderMetadata(self.PROVIDER_BASEURL,
                                             self.PROVIDER_BASEURL + '/auth',
                                             self.PROVIDER_BASEURL + '/jwks',
                                             registration_endpoint=registration_endpoint)
        provider_configurations = {
            self.PROVIDER_NAME: ProviderConfiguration(provider_metadata=provider_metadata,
                                                      client_registration_info=ClientRegistrationInfo(
                                                          redirect_uris=[f'https://{self.CLIENT_DOMAIN}/redirect',
                                                                         f'https://{self.CLIENT_DOMAIN}/redirect2'],
                                                          post_logout_redirect_uris=post_logout_redirect_uris
                                                      ))
        }
        authn = OIDCAuthentication(provider_configurations)
        authn.init_app(self.app)

        # register logout view to force 'post_logout_redirect_uris' to be included in registration request
        logout_view_mock = self.get_view_mock()
        self.app.add_url_rule('/logout', view_func=authn.oidc_logout(logout_view_mock))

        expected_post_logout_redirect_uris = post_logout_redirect_uris if post_logout_redirect_uris else \
            [f'http://{self.CLIENT_DOMAIN}/logout']
        responses.add(responses.POST, registration_endpoint, json={
            'client_id': 'client1', 'client_secret': 'secret1',
            'redirect_uris': [f'https://{self.CLIENT_DOMAIN}/redirect',
                              f'https://{self.CLIENT_DOMAIN}/redirect2'],
            'post_logout_redirect_uris': expected_post_logout_redirect_uris})
        view_mock = self.get_view_mock()
        with self.app.test_request_context('/'):
            auth_redirect = authn.oidc_auth(self.PROVIDER_NAME)(view_mock)()

        self.assert_auth_redirect(auth_redirect)

        registration_request = json.loads(responses.calls[0].request.body)
        expected_registration_request = {'application_type': 'web', 'response_types': ['code'],
                                         'redirect_uris': [f'https://{self.CLIENT_DOMAIN}/redirect',
                                                           f'https://{self.CLIENT_DOMAIN}/redirect2'],
                                         'post_logout_redirect_uris': expected_post_logout_redirect_uris,
                                         'grant_types': ['authorization_code']}
        assert registration_request == expected_registration_request

    @responses.activate
    def test_register_client_should_add_redirect_uri_if_not_passed(self):
        registration_endpoint = self.PROVIDER_BASEURL + '/register'
        provider_metadata = ProviderMetadata(self.PROVIDER_BASEURL,
                                             self.PROVIDER_BASEURL + '/auth',
                                             self.PROVIDER_BASEURL + '/jwks',
                                             registration_endpoint=registration_endpoint)
        post_logout_redirect_uris = [f'https://{self.CLIENT_DOMAIN}/logout']
        provider_configurations = {
            self.PROVIDER_NAME: ProviderConfiguration(
                provider_metadata=provider_metadata,
                client_registration_info=ClientRegistrationInfo(
                    post_logout_redirect_uris=post_logout_redirect_uris
                ))
        }
        authn = OIDCAuthentication(provider_configurations)
        authn.init_app(self.app)

        # register logout view to force 'post_logout_redirect_uris' to be included in registration request
        logout_view_mock = self.get_view_mock()
        self.app.add_url_rule('/logout', view_func=logout_view_mock)
        authn.oidc_logout(logout_view_mock)

        redirect_uris = [f'http://{self.CLIENT_DOMAIN}/redirect_uri']
        responses.add(responses.POST, registration_endpoint, json={
            'client_id': 'client1', 'client_secret': 'secret1',
            'redirect_uris': redirect_uris})
        view_mock = self.get_view_mock()
        with self.app.test_request_context('/'):
            auth_redirect = authn.oidc_auth(self.PROVIDER_NAME)(view_mock)()

        self.assert_auth_redirect(auth_redirect)

        registration_request = json.loads(responses.calls[0].request.body)
        expected_registration_request = {'application_type': 'web', 'response_types': ['code'],
                                         'redirect_uris': redirect_uris,
                                         'post_logout_redirect_uris': post_logout_redirect_uris,
                                         'grant_types': ['authorization_code']}
        assert registration_request == expected_registration_request

    @responses.activate
    def test_register_client_should_return_empty_post_logout_redirect_uris_if_logout_view_not_exist(self):
        registration_endpoint = self.PROVIDER_BASEURL + '/register'
        provider_metadata = ProviderMetadata(self.PROVIDER_BASEURL,
                                             self.PROVIDER_BASEURL + '/auth',
                                             self.PROVIDER_BASEURL + '/jwks',
                                             registration_endpoint=registration_endpoint)
        provider_configurations = {
            self.PROVIDER_NAME: ProviderConfiguration(
                provider_metadata=provider_metadata,
                client_registration_info=ClientRegistrationInfo())
        }
        authn = OIDCAuthentication(provider_configurations)
        authn.init_app(self.app)

        redirect_uris = [f'http://{self.CLIENT_DOMAIN}/redirect_uri']
        responses.add(responses.POST, registration_endpoint, json={
            'client_id': 'client1', 'client_secret': 'secret1',
            'redirect_uris': redirect_uris})
        view_mock = self.get_view_mock()
        with self.app.test_request_context('/'):
            auth_redirect = authn.oidc_auth(self.PROVIDER_NAME)(view_mock)()

        self.assert_auth_redirect(auth_redirect)
        assert authn.clients[self.PROVIDER_NAME]._provider_configuration._client_metadata.get(
            'post_logout_redirect_uris') is None

        registration_request = json.loads(responses.calls[0].request.body)
        expected_registration_request = {'application_type': 'web', 'response_types': ['code'],
                                         'redirect_uris': redirect_uris,
                                         'grant_types': ['authorization_code']}
        assert registration_request == expected_registration_request

    @patch('time.time')
    @patch('oic.utils.time_util.utc_time_sans_frac')  # used internally by pyoidc when verifying ID Token
    @responses.activate
    def test_handle_authentication_response(self, time_mock, utc_time_sans_frac_mock):
        # freeze time since ID Token validation includes expiration timestamps
        timestamp = time.mktime(datetime(2017, 1, 1).timetuple())
        time_mock.return_value = timestamp
        utc_time_sans_frac_mock.return_value = int(timestamp)

        # mock token response
        user_id = 'user1'
        exp_time = 10
        nonce = 'test_nonce'
        id_token_claims = {
            'iss': self.PROVIDER_BASEURL,
            'aud': [self.CLIENT_ID],
            'sub': user_id,
            'exp': int(timestamp) + exp_time,
            'iat': int(timestamp),
            'nonce': nonce
        }
        id_token_jwt, id_token_signing_key = signed_id_token(id_token_claims)
        access_token = 'test_access_token'
        expires_in = 3600
        token_response = {
            'access_token': access_token,
            'expires_in': expires_in,
            'token_type': 'Bearer',
            'id_token': id_token_jwt
        }
        token_endpoint = self.PROVIDER_BASEURL + '/token'
        responses.add(responses.POST, token_endpoint, json=token_response)
        responses.add(responses.GET,
                      self.PROVIDER_BASEURL + '/jwks',
                      json={'keys': [id_token_signing_key.serialize()]})

        # mock userinfo response
        userinfo = {'sub': user_id, 'name': 'Test User'}
        userinfo_endpoint = self.PROVIDER_BASEURL + '/userinfo'
        responses.add(responses.GET, userinfo_endpoint, json=userinfo)

        authn = self.init_app(provider_metadata_extras={'token_endpoint': token_endpoint,
                                                        'userinfo_endpoint': userinfo_endpoint})
        state = 'test_state'
        with self.app.test_request_context(f'/redirect_uri?state={state}&code=test'):
            UserSession(flask.session, self.PROVIDER_NAME)
            flask.session['destination'] = '/'
            flask.session['auth_request'] = json.dumps({'state': state, 'nonce': nonce})
            authn._handle_authentication_response()
            session = UserSession(flask.session)
            assert session.access_token == access_token
            assert session.access_token_expires_at == int(timestamp) + expires_in
            assert session.id_token == id_token_claims
            assert session.id_token_jwt == id_token_jwt
            assert session.userinfo == userinfo

    @patch('time.time')
    @patch('oic.utils.time_util.utc_time_sans_frac')  # used internally by pyoidc when verifying ID Token
    @responses.activate
    def test_handle_implicit_authentication_response(self, time_mock, utc_time_sans_frac_mock):
        # freeze time since ID Token validation includes expiration timestamps
        timestamp = time.mktime(datetime(2017, 1, 1).timetuple())
        time_mock.return_value = timestamp
        utc_time_sans_frac_mock.return_value = int(timestamp)

        # mock auth response
        access_token = 'test_access_token'
        user_id = 'user1'
        exp_time = 10
        nonce = 'test_nonce'
        id_token_claims = {
            'iss': self.PROVIDER_BASEURL,
            'aud': [self.CLIENT_ID],
            'sub': user_id,
            'exp': int(timestamp) + exp_time,
            'iat': int(timestamp),
            'nonce': nonce,
            'at_hash': jws.left_hash(access_token)
        }
        id_token_jwt, id_token_signing_key = signed_id_token(id_token_claims)

        responses.add(responses.GET,
                      self.PROVIDER_BASEURL + '/jwks',
                      json={'keys': [id_token_signing_key.serialize()]})

        # mock userinfo response
        userinfo = {'sub': user_id, 'name': 'Test User'}
        userinfo_endpoint = self.PROVIDER_BASEURL + '/userinfo'
        responses.add(responses.GET, userinfo_endpoint, json=userinfo)

        self.init_app(provider_metadata_extras={'userinfo_endpoint': userinfo_endpoint})
        state = 'test_state'
        auth_response = AuthorizationResponse(
            **{'state': state, 'access_token': access_token, 'token_type': 'Bearer', 'id_token': id_token_jwt})

        with self.app.test_client() as client:
            with client.session_transaction() as session:
                UserSession(session, self.PROVIDER_NAME)
                session['destination'] = '/'
                session['auth_request'] = json.dumps({'state': state, 'nonce': nonce})
                session['fragment_encoded_response'] = True
            client.get(f'/redirect_uri#{auth_response.to_urlencoded()}')
            assert 'auth_request' in session  # stored auth_request should not have been removed yet

            # fake the POST request from the 'parse_fragment.html' template
            resp = client.post('/redirect_uri', data=auth_response.to_dict())
            user_session = UserSession(flask.session)
            assert user_session.access_token == access_token
            assert user_session.id_token == id_token_claims
            assert user_session.id_token_jwt == id_token_jwt
            assert user_session.userinfo == userinfo
            assert 'auth_request' not in flask.session  # stored auth_request should have been removed now
            assert resp.data.decode('utf-8') == '/'  # final redirect back to the protected endpoint

    def test_handle_authentication_response_POST(self):
        access_token = 'test_access_token'
        state = 'test_state'

        authn = self.init_app()
        auth_response = AuthorizationResponse(**{'state': state, 'token_type': 'Bearer', 'access_token': access_token})

        with self.app.test_request_context('/redirect_uri',
                                           method='POST',
                                           data=auth_response.to_dict(),
                                           mimetype='application/x-www-form-urlencoded'):
            UserSession(flask.session, self.PROVIDER_NAME)
            flask.session['destination'] = '/test'
            flask.session['auth_request'] = json.dumps({'state': state, 'nonce': 'test_nonce'})
            response = authn._handle_authentication_response()
            session = UserSession(flask.session)
            assert session.access_token == access_token
            assert response == '/test'

    def test_handle_error_response_POST(self):
        state = 'test_state'

        authn = self.init_app()
        error_resp = {'state': state, 'error': 'invalid_request', 'error_description': 'test error'}

        with self.app.test_request_context('/redirect_uri',
                                           method='POST',
                                           data=error_resp,
                                           mimetype='application/x-www-form-urlencoded'):
            UserSession(flask.session, self.PROVIDER_NAME)
            flask.session['auth_request'] = json.dumps({'state': state, 'nonce': 'test_nonce'})
            response = authn._handle_authentication_response()
            assert flask.session['error'] == error_resp
            assert response == '/redirect_uri?error=1'

    def test_handle_authentication_response_without_initialised_session(self):
        authn = self.init_app()

        with self.app.test_request_context('/redirect_uri?state=test-state&code=test'):
            response = authn._handle_authentication_response()
            assert response == 'Something went wrong with the authentication, please try to login again.'

            # with error view configured, error object should be sent to it instead
            error_view_mock = self.get_view_mock()
            authn.error_view(error_view_mock)
            result = authn._handle_authentication_response()
            self.assert_view_mock(error_view_mock, result)
            error_view_mock.assert_called_with(
                **{'error': 'unsolicited_response', 'error_description': 'No initialised user session.'})

    def test_handle_authentication_response_without_stored_auth_request(self):
        authn = self.init_app()

        with self.app.test_request_context('/redirect_uri?state=test-state&code=test'):
            UserSession(flask.session, self.PROVIDER_NAME)
            flask.session['destination'] = '/test'
            response = authn._handle_authentication_response()
            assert response == 'Something went wrong with the authentication, please try to login again.'

            # with error view configured, error object should be sent to it instead
            error_view_mock = self.get_view_mock()
            authn.error_view(error_view_mock)
            result = authn._handle_authentication_response()
            self.assert_view_mock(error_view_mock, result)
            error_view_mock.assert_called_with(
                **{'error': 'unsolicited_response', 'error_description': 'No authentication request stored.'})

    def test_handle_authentication_response_fragment_encoded(self):
        authn = self.init_app()
        with self.app.test_request_context('/redirect_uri'):
            UserSession(flask.session, self.PROVIDER_NAME)
            flask.session['auth_request'] = json.dumps({'state': 'test_state', 'nonce': 'test_nonce'})
            flask.session['fragment_encoded_response'] = True
            response = authn._handle_authentication_response()
        assert response.startswith('<html>')

    def test_handle_authentication_response_error_message(self):
        authn = self.init_app()
        with self.app.test_request_context('/redirect_uri?error=1'):
            flask.session['error'] = {'error': 'test'}
            response = authn._handle_authentication_response()
        assert response == 'Something went wrong with the authentication, please try to login again.'

    def test_handle_authentication_response_error_message_without_stored_error(self):
        authn = self.init_app()
        with self.app.test_request_context('/redirect_uri?error=1'):
            response = authn._handle_authentication_response()
        assert response == 'Something went wrong.'

    @patch('time.time')
    @patch('oic.utils.time_util.utc_time_sans_frac')  # used internally by pyoidc when verifying ID Token
    @responses.activate
    def test_session_expiration_set_to_configured_lifetime(self, time_mock, utc_time_sans_frac_mock):
        timestamp = time.mktime(datetime(2017, 1, 1).timetuple())
        time_mock.return_value = timestamp
        utc_time_sans_frac_mock.return_value = int(timestamp)

        exp_time = 10
        state = 'test_state'
        nonce = 'test_nonce'
        id_token = IdToken(iss=self.PROVIDER_BASEURL,
                           aud=self.CLIENT_ID,
                           sub='sub1',
                           exp=int(timestamp) + exp_time,
                           iat=int(timestamp),
                           nonce=nonce)
        token_response = {'access_token': 'test', 'token_type': 'Bearer', 'id_token': id_token.to_jwt()}
        token_endpoint = self.PROVIDER_BASEURL + '/token'
        responses.add(responses.POST, token_endpoint, json=token_response)

        session_lifetime = 1234
        self.app.config['PERMANENT_SESSION_LIFETIME'] = session_lifetime
        self.init_app(provider_metadata_extras={'token_endpoint': token_endpoint})

        with self.app.test_client() as client:
            with client.session_transaction() as session:
                UserSession(session, self.PROVIDER_NAME)
                session['destination'] = '/'
                session['auth_request'] = json.dumps({'state': state, 'nonce': nonce, 'response_type': 'code'})
            resp = client.get(f'/redirect_uri?state={state}&code=test')

        cookies = SimpleCookie()
        cookies.load(resp.headers['Set-Cookie'])
        session_cookie_expiration = cookies[self.app.config['SESSION_COOKIE_NAME']]['expires']
        parsed_expiration = datetime.strptime(session_cookie_expiration, '%a, %d %b %Y %H:%M:%S GMT')
        cookie_lifetime = (parsed_expiration - datetime.utcnow()).total_seconds()
        assert cookie_lifetime == pytest.approx(session_lifetime, abs=1)

    def test_logout_redirects_to_provider_if_end_session_endpoint_is_configured(self):
        end_session_endpoint = 'https://provider.example.com/end_session'
        client_metadata = {}

        authn = self.init_app(provider_metadata_extras={'end_session_endpoint': end_session_endpoint},
                              client_metadata_extras=client_metadata)
        logout_view_mock = self.get_view_mock()
        id_token = IdToken(**{'sub': 'sub1', 'nonce': 'nonce'})

        # register logout view
        view_func = authn.oidc_logout(logout_view_mock)
        self.app.add_url_rule('/logout', view_func=view_func)

        with self.app.test_request_context('/logout'):
            UserSession(flask.session, self.PROVIDER_NAME).update(access_token='test_access_token',
                                                                  id_token=id_token.to_dict(),
                                                                  id_token_jwt=id_token.to_jwt(),
                                                                  userinfo={'sub': 'user1'})
            end_session_redirect = view_func()
            # ensure user session has been cleared
            assert all(k not in flask.session for k in UserSession.KEYS)
            parsed_request = dict(parse_qsl(urlparse(end_session_redirect.headers['Location']).query))
            assert parsed_request['state'] == flask.session['end_session_state']

        assert end_session_redirect.status_code == 303
        assert end_session_redirect.location.startswith(end_session_endpoint)
        assert IdToken().from_jwt(parsed_request['id_token_hint']) == id_token

        assert parsed_request['post_logout_redirect_uri'] == f'http://{self.CLIENT_DOMAIN}/logout'
        assert not logout_view_mock.called

    @responses.activate
    def test_multiple_logout_endpoints_are_supported(self):
        end_session_endpoint = 'https://provider.example.com/end_session'
        registration_endpoint = self.PROVIDER_BASEURL + '/register'
        id_token = IdToken(**{'sub': 'sub1', 'nonce': 'nonce'})

        provider_metadata = ProviderMetadata(self.PROVIDER_BASEURL,
                                             self.PROVIDER_BASEURL + '/auth',
                                             self.PROVIDER_BASEURL + '/jwks',
                                             registration_endpoint=registration_endpoint,
                                             end_session_endpoint=end_session_endpoint)
        post_logout_redirect_uris = [f'http://{self.CLIENT_DOMAIN}/logout1',
                                     f'http://{self.CLIENT_DOMAIN}/logout2']
        client_registration_info = ClientRegistrationInfo(post_logout_redirect_uris=post_logout_redirect_uris)
        provider_configurations = {
            self.PROVIDER_NAME: ProviderConfiguration(provider_metadata=provider_metadata,
                                                      client_registration_info=client_registration_info)
        }
        authn = OIDCAuthentication(provider_configurations)
        authn.init_app(self.app)

        # register multiple logout views
        view_func1 = authn.oidc_logout(self.get_view_mock('logout'))
        self.app.add_url_rule('/logout1', 'logout', view_func=view_func1)
        view_func2 = authn.oidc_logout(self.get_view_mock('otherlogout'))
        # register logout view with custom endpoint
        self.app.add_url_rule('/logout2', 'test.otherlogout', view_func=view_func2)

        # verify client registration includes all logout endpoints as 'post_logout_redirect_uris'
        responses.add(responses.POST, registration_endpoint, json={
            'client_id': 'client1', 'client_secret': 'secret1',
            'redirect_uris': [f'http://{self.CLIENT_DOMAIN}/redirect_uri'],
            'post_logout_redirect_uris': post_logout_redirect_uris})
        view_mock = self.get_view_mock()
        with self.app.test_request_context('/'):
            authn.oidc_auth(self.PROVIDER_NAME)(view_mock)()

        registration_request = json.loads(responses.calls[0].request.body)
        expected_registration_request = {'application_type': 'web', 'response_types': ['code'],
                                         'redirect_uris': [f'http://{self.CLIENT_DOMAIN}/redirect_uri'],
                                         'post_logout_redirect_uris': post_logout_redirect_uris,
                                         'grant_types': ['authorization_code']}
        assert registration_request == expected_registration_request

        # verify each logout endpoint can be called
        for endpoint, view_func in [('/logout1', view_func1), ('/logout2', view_func2)]:
            with self.app.test_request_context(endpoint):
                UserSession(flask.session, self.PROVIDER_NAME).update(access_token='test_access_token',
                                                                      id_token=id_token.to_dict(),
                                                                      id_token_jwt=id_token.to_jwt(),
                                                                      userinfo={'sub': 'user1'})
                end_session_redirect = view_func()
                # ensure user session has been cleared
                assert all(k not in flask.session for k in UserSession.KEYS)
                parsed_request = dict(parse_qsl(urlparse(end_session_redirect.headers['Location']).query))
                assert parsed_request['state'] == flask.session['end_session_state']

            assert end_session_redirect.status_code == 303
            assert end_session_redirect.location.startswith(end_session_endpoint)
            assert IdToken().from_jwt(parsed_request['id_token_hint']) == id_token

            expected_post_logout_redirect_uri = f'http://{self.CLIENT_DOMAIN}{endpoint}'
            assert parsed_request['post_logout_redirect_uri'] == expected_post_logout_redirect_uri

    def test_logout_with_missing_end_session_state_fails_gracefully(self):
        end_session_endpoint = 'https://provider.example.com/end_session'
        authn = self.init_app(provider_metadata_extras={'end_session_endpoint': end_session_endpoint})
        id_token = IdToken(**{'sub': 'sub1', 'nonce': 'nonce'})
        logout_view_mock = self.get_view_mock()

        # register logout view
        view_func = authn.oidc_logout(logout_view_mock)
        self.app.add_url_rule('/logout', view_func=view_func)

        with self.app.test_request_context('/logout?state=incorrect'):
            UserSession(flask.session, self.PROVIDER_NAME).update(access_token='test_access_token',
                                                                  id_token=id_token.to_dict(),
                                                                  id_token_jwt=id_token.to_jwt(),
                                                                  userinfo={'sub': 'user1'})
            flask.session.pop('end_session_state', None)  # make sure there's no 'end_session_state'
            logout_result = authn.oidc_logout(logout_view_mock)()

        self.assert_view_mock(logout_view_mock, logout_result)

    def test_logout_handles_provider_without_end_session_endpoint(self):
        authn = self.init_app()
        id_token = IdToken(**{'sub': 'sub1', 'nonce': 'nonce'})
        logout_view_mock = self.get_view_mock()
        with self.app.test_request_context('/logout'):
            UserSession(flask.session, self.PROVIDER_NAME).update(access_token='test_access_token',
                                                                  id_token=id_token.to_dict(),
                                                                  id_token_jwt=id_token.to_jwt(),
                                                                  userinfo={'sub': 'user1'})

            logout_result = authn.oidc_logout(logout_view_mock)()
            assert all(k not in flask.session for k in UserSession.KEYS)

        self.assert_view_mock(logout_view_mock, logout_result)

    def test_logout_handles_redirect_back_from_provider(self):
        authn = self.init_app()
        logout_view_mock = self.get_view_mock()
        state = 'end_session_123'
        with self.app.test_request_context(f'/logout?state={state}'):
            flask.session['end_session_state'] = state
            result = authn.oidc_logout(logout_view_mock)()
            assert 'end_session_state' not in flask.session

        self.assert_view_mock(logout_view_mock, result)

    def test_logout_handles_redirect_back_from_provider_with_incorrect_state(self, caplog):
        authn = self.init_app()
        logout_view_mock = self.get_view_mock()
        state = 'some_state'
        with self.app.test_request_context(f'/logout?state={state}'):
            flask.session['end_session_state'] = 'other_state'
            result = authn.oidc_logout(logout_view_mock)()
            assert 'end_session_state' not in flask.session

        self.assert_view_mock(logout_view_mock, result)
        assert caplog.record_tuples[-1] == ('flask_pyoidc.flask_pyoidc',
                                            logging.ERROR,
                                            f"Got unexpected state '{state}' after logout redirect.")

    def test_logout_handles_no_user_session(self):
        authn = self.init_app()
        logout_view_mock = self.get_view_mock()
        with self.app.test_request_context('/logout'):
            result = authn.oidc_logout(logout_view_mock)()

        self.assert_view_mock(logout_view_mock, result)

    def test_authentication_error_response_calls_to_error_view_if_set(self):
        state = 'test_tate'
        error_response = {'error': 'invalid_request', 'error_description': 'test error'}
        authn = self.init_app()
        error_view_mock = self.get_view_mock()
        authn.error_view(error_view_mock)
        with self.app.test_request_context(f'/redirect_uri?{urlencode(error_response)}&state={state}'):
            UserSession(flask.session, self.PROVIDER_NAME)
            flask.session['auth_request'] = json.dumps({'state': state, 'nonce': 'test_nonce'})
            result = authn._handle_authentication_response()

        self.assert_view_mock(error_view_mock, result)
        error_view_mock.assert_called_with(**error_response)

    def test_authentication_error_response_returns_default_error_if_no_error_view_set(self):
        state = 'test_tate'
        error_response = {'error': 'invalid_request', 'error_description': 'test error', 'state': state}
        config = {
            'provider_configuration_info': {'issuer': self.PROVIDER_BASEURL},
            'client_registration_info': {'client_id': 'abc', 'client_secret': 'foo'}
        }
        authn = self.init_app(config)
        with self.app.test_request_context(f'/redirect_uri?{urlencode(error_response)}'):
            UserSession(flask.session, self.PROVIDER_NAME)
            flask.session['state'] = state
            flask.session['nonce'] = 'test_nonce'
            response = authn._handle_authentication_response()
        assert response == "Something went wrong with the authentication, please try to login again."

    @responses.activate
    def test_token_error_response_calls_to_error_view_if_set(self):
        token_endpoint = self.PROVIDER_BASEURL + '/token'
        error_response = {'error': 'invalid_request', 'error_description': 'test error'}
        responses.add(responses.POST, token_endpoint, json=error_response)

        authn = self.init_app(provider_metadata_extras={'token_endpoint': token_endpoint})
        error_view_mock = self.get_view_mock()
        authn.error_view(error_view_mock)
        state = 'test_tate'
        with self.app.test_request_context(f'/redirect_uri?code=foo&state={state}'):
            UserSession(flask.session, self.PROVIDER_NAME)
            flask.session['auth_request'] = json.dumps({'state': state, 'nonce': 'test_nonce'})
            result = authn._handle_authentication_response()

        self.assert_view_mock(error_view_mock, result)
        error_view_mock.assert_called_with(**error_response)

    @responses.activate
    def test_token_error_response_returns_default_error_if_no_error_view_set(self):
        token_endpoint = self.PROVIDER_BASEURL + '/token'
        state = 'test_tate'
        error_response = {'error': 'invalid_request', 'error_description': 'test error', 'state': state}
        responses.add(responses.POST, token_endpoint, json=error_response)

        authn = self.init_app(provider_metadata_extras={'token_endpoint': token_endpoint})
        with self.app.test_request_context('/redirect_uri?code=foo&state=' + state):
            UserSession(flask.session, self.PROVIDER_NAME)
            flask.session['state'] = state
            flask.session['nonce'] = 'test_nonce'
            response = authn._handle_authentication_response()
        assert response == "Something went wrong with the authentication, please try to login again."

    def test_using_unknown_provider_name_should_raise_exception(self):
        with pytest.raises(ValueError) as exc_info:
            self.init_app().oidc_auth('unknown')
        assert 'unknown' in str(exc_info.value)

    def test_should_not_refresh_if_no_user_session(self):
        with self.app.test_request_context('/foo'):
            assert self.init_app().valid_access_token() is None

    @responses.activate
    def test_should_refresh_expired_access_token(self):
        token_endpoint = self.PROVIDER_BASEURL + '/token'
        authn = self.init_app(provider_metadata_extras={'token_endpoint': token_endpoint})

        token_response = {
            'access_token': 'new-access-token',
            'expires_in': 3600,
            'token_type': 'Bearer',
            'refresh_token': 'new-refresh-token'
        }
        responses.add(responses.POST, token_endpoint, json=token_response)

        with self.app.test_request_context('/foo'):
            session = UserSession(flask.session, self.PROVIDER_NAME)
            session.update(expires_in=-10, refresh_token='refresh-token')
            assert authn.valid_access_token() == token_response['access_token']
            assert session.access_token == token_response['access_token']
            assert session.refresh_token == token_response['refresh_token']

    def test_should_not_refresh_still_valid_access_token(self):
        authn = self.init_app()

        access_token = 'access_token'
        with self.app.test_request_context('/foo'):
            session = UserSession(flask.session, self.PROVIDER_NAME)
            session.update(access_token=access_token, expires_in=10, refresh_token='refresh-token')
            assert authn.valid_access_token() == access_token

    @responses.activate
    def test_should_refresh_still_valid_access_token_if_forced(self):
        token_endpoint = self.PROVIDER_BASEURL + '/token'
        authn = self.init_app(provider_metadata_extras={'token_endpoint': token_endpoint})

        token_response = {
            'access_token': 'new-access-token',
            'expires_in': 3600,
            'token_type': 'Bearer',
            'refresh_token': 'new-refresh-token'
        }
        responses.add(responses.POST, token_endpoint, json=token_response)

        with self.app.test_request_context('/foo'):
            session = UserSession(flask.session, self.PROVIDER_NAME)
            session.update(expires_in=10, refresh_token='refresh-token')
            assert authn.valid_access_token(force_refresh=True) == token_response['access_token']
            assert session.access_token == token_response['access_token']
            assert session.refresh_token == token_response['refresh_token']

    def test_should_not_refresh_without_refresh_token(self):
        authn = self.init_app()

        with self.app.test_request_context('/foo'):
            session = UserSession(flask.session, self.PROVIDER_NAME)
            session.update(expires_in=-10)
            assert authn.valid_access_token() is None

    def test_should_not_refresh_access_token_without_expiry(self):
        authn = self.init_app()

        access_token = 'access_token'
        with self.app.test_request_context('/foo'):
            session = UserSession(flask.session, self.PROVIDER_NAME)
            session.update(access_token=access_token, refresh_token='refresh-token')
            assert authn.valid_access_token() == access_token

    @responses.activate
    def test_should_return_None_if_token_refresh_request_fails(self):
        token_endpoint = self.PROVIDER_BASEURL + '/token'
        authn = self.init_app(provider_metadata_extras={'token_endpoint': token_endpoint})

        token_response = {
            'error': 'invalid_grant',
            'error_description': 'The refresh token is invalid'
        }
        responses.add(responses.POST, token_endpoint, json=token_response)

        access_token = 'access_token'
        with self.app.test_request_context('/foo'):
            session = UserSession(flask.session, self.PROVIDER_NAME)
            session.update(access_token=access_token, expires_in=-10, refresh_token='refresh-token')
            assert authn.valid_access_token(force_refresh=True) is None
            assert session.access_token == access_token

    def test_should_check_for_authorization_header(self):

        authn = self.init_app()
        with self.app.test_request_context('/'):
            assert not authn._check_authorization_header(flask.request)
            flask.request.headers = {
                'Authorization': 'Bearer access_token'
            }
            assert authn._check_authorization_header(flask.request)

    def test_should_parse_access_token_from_request_header(self):

        authn = self.init_app()
        with self.app.test_request_context('/'):
            flask.request.headers = {
                'Authorization': 'Bearer access_token'
            }
            assert authn._parse_access_token(flask.request) == 'access_token'

    @responses.activate
    def test_introspect_token_should_return_none_if_invalid_access_token(self):

        introspection_endpoint = f'{self.PROVIDER_BASEURL}/token/introspect'
        authn = self.init_app(provider_metadata_extras={
            'introspection_endpoint': introspection_endpoint})
        with self.app.test_request_context('/'):
            flask.request.headers = {
                'Authorization': 'Bearer access_token'
            }
            responses.add(responses.POST, introspection_endpoint,
                          json={'active': False})
            assert authn.introspect_token(
                flask.request, authn.clients[self.PROVIDER_NAME]) is None

    @responses.activate
    def test_introspect_token_should_return_none_if_client_id_not_in_audience(self):

        introspection_endpoint = f'{self.PROVIDER_BASEURL}/token/introspect'
        authn = self.init_app(provider_metadata_extras={
            'introspection_endpoint': introspection_endpoint})
        with self.app.test_request_context('/'):
            flask.request.headers = {
                'Authorization': 'Bearer access_token'
            }
            token_introspection_response = {
                'active': True,
                'aud': ['admin', 'user']
            }
            responses.add(responses.POST, introspection_endpoint,
                          json=token_introspection_response)
            assert authn.introspect_token(
                flask.request, authn.clients[self.PROVIDER_NAME]) is None

    @responses.activate
    def test_introspect_token_should_return_none_if_required_scopes_not_permitted(self):

        introspection_endpoint = f'{self.PROVIDER_BASEURL}/token/introspect'
        authn = self.init_app(provider_metadata_extras={
            'introspection_endpoint': introspection_endpoint})
        with self.app.test_request_context('/'):
            flask.request.headers = {
                'Authorization': 'Bearer access_token'
            }
            token_introspection_response = {
                'active': True,
                'aud': ['admin', 'user', self.CLIENT_ID],
                'scope': ['read', 'write']
            }
            responses.add(responses.POST, introspection_endpoint,
                          json=token_introspection_response)
            assert authn.introspect_token(
                flask.request, authn.clients[self.PROVIDER_NAME],
                scopes=['read', 'write', 'delete']) is None

    @responses.activate
    def test_introspect_token_should_return_introspection_result_if_valid_access_token(self):

        introspection_endpoint = f'{self.PROVIDER_BASEURL}/token/introspect'
        authn = self.init_app(provider_metadata_extras={
            'introspection_endpoint': introspection_endpoint})
        with self.app.test_request_context('/'):
            flask.request.headers = {
                'Authorization': 'Bearer access_token'
            }
            token_introspection_response = {
                'active': True,
                'aud': ['admin', 'user', self.CLIENT_ID],
                'scope': 'read write delete',
                'client_id': self.CLIENT_ID
            }
            responses.add(responses.POST, introspection_endpoint,
                          json=token_introspection_response)
            introspection_result = authn.introspect_token(
                flask.request, authn.clients[self.PROVIDER_NAME])
            assert token_introspection_response == introspection_result.to_dict()

    def test_token_auth_should_raise_unauthorized_if_authorization_missing(self):

        authn = self.init_app()
        view_mock = self.get_view_mock()
        with self.app.test_request_context('/'):
            with pytest.raises(Unauthorized):
                authn.token_auth(self.PROVIDER_NAME)(view_mock)()

    @responses.activate
    def test_token_auth_should_run_view_function_if_valid_token(self):

        introspection_endpoint = f'{self.PROVIDER_BASEURL}/token/introspect'
        authn = self.init_app(provider_metadata_extras={
            'introspection_endpoint': introspection_endpoint})
        view_mock = self.get_view_mock()
        introspection_endpoint = f'{self.PROVIDER_BASEURL}/token/introspect'
        token_introspection_response = {
            'active': True,
            'aud': ['admin', 'user', self.CLIENT_ID],
            'scope': 'read write delete',
            'client_id': self.CLIENT_ID
        }
        responses.add(responses.POST, introspection_endpoint,
                      json=token_introspection_response)
        with self.app.test_request_context('/'):
            flask.request.headers = {
                'Authorization': 'Bearer access_token'
            }
            authn.token_auth(self.PROVIDER_NAME,
                             scopes_required=['read', 'write'])(view_mock)()
            assert view_mock.called
            assert flask.g.current_token_identity == token_introspection_response

    @responses.activate
    def test_token_auth_should_raise_forbidden_if_invalid_token(self):

        introspection_endpoint = f'{self.PROVIDER_BASEURL}/token/introspect'
        authn = self.init_app(provider_metadata_extras={
            'introspection_endpoint': introspection_endpoint})
        view_mock = self.get_view_mock()
        introspection_endpoint = f'{self.PROVIDER_BASEURL}/token/introspect'
        token_introspection_response = {
            'active': False,
            'aud': ['admin', 'user', self.CLIENT_ID],
            'scope': 'read write delete',
            'client_id': self.CLIENT_ID
        }
        responses.add(responses.POST, introspection_endpoint,
                      json=token_introspection_response)
        with self.app.test_request_context('/'):
            flask.request.headers = {
                'Authorization': 'Bearer access_token'
            }
            with pytest.raises(Forbidden):
                authn.token_auth(
                    self.PROVIDER_NAME,
                    scopes_required=['read', 'write'])(view_mock)()

    @responses.activate
    def test_access_control_should_fallback_to_oidc_auth_on_401(self):

        authn = self.init_app()
        view_mock = self.get_view_mock()
        with self.app.test_request_context('/'):
            auth_fallback = authn.access_control(
                self.PROVIDER_NAME)(view_mock)()
            self.assert_auth_redirect(auth_fallback)
            assert not view_mock.called

    @responses.activate
    def test_access_control_should_abort_request_if_invalid_token(self):

        introspection_endpoint = f'{self.PROVIDER_BASEURL}/token/introspect'
        authn = self.init_app(provider_metadata_extras={
            'introspection_endpoint': introspection_endpoint})
        view_mock = self.get_view_mock()
        introspection_endpoint = f'{self.PROVIDER_BASEURL}/token/introspect'
        token_introspection_response = {
            'active': False,
            'aud': ['admin', 'user', self.CLIENT_ID],
            'scope': 'read write delete',
            'client_id': self.CLIENT_ID
        }
        responses.add(responses.POST, introspection_endpoint,
                      json=token_introspection_response)
        with self.app.test_request_context('/'):
            flask.request.headers = {
                'Authorization': 'Bearer access_token'
            }
            with pytest.raises(Forbidden):
                authn.access_control(
                    self.PROVIDER_NAME,
                    scopes_required=['read', 'write'])(view_mock)()

    @responses.activate
    def test_access_control_should_run_view_function_if_valid_token(self):

        introspection_endpoint = f'{self.PROVIDER_BASEURL}/token/introspect'
        authn = self.init_app(provider_metadata_extras={
            'introspection_endpoint': introspection_endpoint})
        view_mock = self.get_view_mock()
        introspection_endpoint = f'{self.PROVIDER_BASEURL}/token/introspect'
        token_introspection_response = {
            'active': True,
            'aud': ['admin', 'user', self.CLIENT_ID],
            'scope': 'read write delete',
            'client_id': self.CLIENT_ID
        }
        responses.add(responses.POST, introspection_endpoint,
                      json=token_introspection_response)
        with self.app.test_request_context('/'):
            flask.request.headers = {
                'Authorization': 'Bearer access_token'
            }
            authn.access_control(
                self.PROVIDER_NAME,
                scopes_required=['read', 'write'])(view_mock)()
            assert view_mock.called
            assert flask.g.current_token_identity == token_introspection_response

    def test_get_url_for_logout_view_should_raise_build_error_if_mounted_under_custom_endpoint(self):
        authn = self.init_app()
        logout_view_mock = self.get_view_mock()
        self.app.add_url_rule('/logout', endpoint='test.logout', view_func=authn.oidc_logout(logout_view_mock))

        with self.app.test_request_context('/'):
            with pytest.raises(BuildError):
                authn._get_urls_for_logout_views()

    @patch('time.time')
    @patch('oic.utils.time_util.utc_time_sans_frac')  # used internally by pyoidc when verifying ID Token
    @responses.activate
    def test_oidc_clock_skew_passed(self, time_mock, utc_time_sans_frac_mock):
        # freeze time since ID Token validation includes expiration timestamps
        timestamp = time.mktime(datetime(2017, 1, 1).timetuple())
        time_mock.return_value = timestamp
        utc_time_sans_frac_mock.return_value = int(timestamp)

        # mock token response
        user_id = 'user1'
        skew = 10
        exp_time = 10
        nonce = 'test_nonce'
        id_token_claims = {
            'iss': self.PROVIDER_BASEURL,
            'aud': [self.CLIENT_ID],
            'sub': user_id,
            'exp': int(timestamp) + exp_time + skew,
            'iat': int(timestamp) + skew,
            'nonce': nonce
        }

        id_token_jwt, id_token_signing_key = signed_id_token(id_token_claims)
        access_token = 'test_access_token'
        expires_in = 3600
        token_response = {
            'access_token': access_token,
            'expires_in': expires_in,
            'token_type': 'Bearer',
            'id_token': id_token_jwt
        }
        token_endpoint = self.PROVIDER_BASEURL + '/token'
        responses.add(responses.POST, token_endpoint, json=token_response)
        responses.add(responses.GET,
                      self.PROVIDER_BASEURL + '/jwks',
                      json={'keys': [id_token_signing_key.serialize()]})

        # mock userinfo response
        userinfo = {'sub': user_id, 'name': 'Test User'}
        userinfo_endpoint = self.PROVIDER_BASEURL + '/userinfo'
        responses.add(responses.GET, userinfo_endpoint, json=userinfo)

        authn = self.init_app(provider_metadata_extras={'token_endpoint': token_endpoint,
                                                        'userinfo_endpoint': userinfo_endpoint})
        state = 'test_state'
        with self.app.test_request_context('/redirect_uri?state={}&code=test'.format(state)):
            UserSession(flask.session, self.PROVIDER_NAME)
            flask.session['destination'] = '/'
            flask.session['auth_request'] = json.dumps({'state': state, 'nonce': nonce})
            authn._handle_authentication_response()
            session = UserSession(flask.session)
            assert session.access_token == access_token
            assert session.access_token_expires_at == int(timestamp) + expires_in
            assert session.id_token == id_token_claims
            assert session.id_token_jwt == id_token_jwt
            assert session.userinfo == userinfo



#   Copyright 2018 Samuel Gulliksson
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0

#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.


import functools
import json
import logging
import time
from typing import Optional
from urllib.parse import parse_qsl

import flask
import importlib_resources
from flask import current_app, g
from flask.helpers import url_for
from oic import rndstr
from oic.extension.message import TokenIntrospectionResponse
from oic.oic import AuthorizationRequest
from oic.oic.message import EndSessionRequest
from werkzeug.exceptions import Forbidden, Unauthorized
from werkzeug.local import LocalProxy
from werkzeug.routing import BuildError
from werkzeug.utils import redirect

from .auth_response_handler import AuthResponseProcessError, AuthResponseHandler, AuthResponseErrorResponseError
from .pyoidc_facade import PyoidcFacade
from .redirect_uri_config import RedirectUriConfig
from .user_session import UninitialisedSession, UserSession

logger = logging.getLogger(__name__)


class OIDCAuthentication:
    """
    OIDCAuthentication object for Flask extension.
    """

    def __init__(self, provider_configurations, app=None,
                 redirect_uri_config=None):
        """
        Args:
            provider_configurations (Mapping[str, ProviderConfiguration]):
                provider configurations by name
            app (flask.app.Flask): optional Flask app
            redirect_uri_config (RedirectUriConfig): optional redirect URI config to use instead of
                'OIDC_REDIRECT_URI' config parameter.
        """
        self._provider_configurations = provider_configurations

        self.clients = None
        self._logout_views = []
        self._error_view = None
        # current_token_identity proxy to obtain user info whose token was
        # passed in the request. It is available until current request only and
        # is destroyed between the requests. The value is set by token_auth
        # decorator.
        self.current_token_identity = LocalProxy(lambda: getattr(
            g, 'current_token_identity', None))
        self._redirect_uri_config = redirect_uri_config

        if app:
            self.init_app(app)

    def init_app(self, app):
        if not self._redirect_uri_config:
            self._redirect_uri_config = RedirectUriConfig.from_config(app.config)

        # setup redirect_uri as a flask route
        app.add_url_rule('/' + self._redirect_uri_config.endpoint,
                         self._redirect_uri_config.endpoint,
                         self._handle_authentication_response,
                         methods=['GET', 'POST'])

        # dynamically add the Flask redirect uri to the client info
        self.clients = {
            name: PyoidcFacade(configuration, self._redirect_uri_config.full_uri)
            for (name, configuration) in self._provider_configurations.items()
        }

    def _get_urls_for_logout_views(self):
        try:
            return [url_for(view.__name__, _external=True) for view in self._logout_views]
        except BuildError:
            logger.error('could not build url for logout view, it might be mounted under a custom endpoint')
            raise

    def _register_client(self, client):
        if not client._provider_configuration._client_registration_info.get('redirect_uris'):
            client._provider_configuration._client_registration_info[
                'redirect_uris'] = [self._redirect_uri_config.full_uri]
        post_logout_redirect_uris = client._provider_configuration._client_registration_info.get(
            'post_logout_redirect_uris')
        if not post_logout_redirect_uris:
            client._provider_configuration._client_registration_info[
                'post_logout_redirect_uris'] = self._get_urls_for_logout_views()
        logger.debug(
            f'''registering with post_logout_redirect_uris = {
                client._provider_configuration._client_registration_info[
                    'post_logout_redirect_uris']}''')
        client.register()

    def _authenticate(self, client, interactive=True):
        if not client.is_registered():
            self._register_client(client)

        flask.session['destination'] = flask.request.url

        # Use silent authentication for session refresh
        # This will not show login prompt to the user
        extra_auth_params = {}
        if not interactive:
            extra_auth_params['prompt'] = 'none'

        auth_req = client.authentication_request(state=rndstr(),
                                                 nonce=rndstr(),
                                                 extra_auth_params=extra_auth_params)
        flask.session['auth_request'] = auth_req.to_json()
        login_url = client.login_url(auth_req)

        auth_params = dict(parse_qsl(login_url.split('?')[1]))
        flask.session['fragment_encoded_response'] = AuthResponseHandler.expect_fragment_encoded_response(auth_params)
        return redirect(login_url)

    def _handle_authentication_response(self):
        has_error = flask.request.args.get('error', False, lambda x: bool(int(x)))
        if has_error:
            if 'error' in flask.session:
                return self._show_error_response(flask.session.pop('error'))
            return 'Something went wrong.'

        try:
            session = UserSession(flask.session)
        except UninitialisedSession:
            return self._handle_error_response({'error': 'unsolicited_response', 'error_description': 'No initialised user session.'})

        if flask.session.pop('fragment_encoded_response', False):
            return importlib_resources.read_binary('flask_pyoidc', 'parse_fragment.html').decode('utf-8')

        if 'auth_request' not in flask.session:
            return self._handle_error_response({'error': 'unsolicited_response', 'error_description': 'No authentication request stored.'})
        auth_request = AuthorizationRequest().from_json(flask.session.pop('auth_request'))

        is_processing_fragment_encoded_response = flask.request.method == 'POST'
        if is_processing_fragment_encoded_response:
            auth_resp = flask.request.form
        else:
            auth_resp = flask.request.args

        client = self.clients[session.current_provider]

        authn_resp = client.parse_authentication_response(auth_resp)
        logger.debug('received authentication response: %s', authn_resp.to_json())

        try:
            extra_token_args = {}
            if 'OIDC_CLOCK_SKEW' in current_app.config:
                extra_token_args['skew'] = current_app.config['OIDC_CLOCK_SKEW']
            result = AuthResponseHandler(client).process_auth_response(authn_resp, auth_request, extra_token_args)
        except AuthResponseErrorResponseError as ex:
            return self._handle_error_response(ex.error_response, is_processing_fragment_encoded_response)
        except AuthResponseProcessError as ex:
            return self._handle_error_response({'error': 'unexpected_error', 'error_description': str(ex)},
                                               is_processing_fragment_encoded_response)

        if current_app.config.get('OIDC_SESSION_PERMANENT', True):
            flask.session.permanent = True

        UserSession(flask.session).update(access_token=result.access_token,
                                          expires_in=result.expires_in,
                                          id_token=result.id_token_claims,
                                          id_token_jwt=result.id_token_jwt,
                                          userinfo=result.userinfo_claims,
                                          refresh_token=result.refresh_token)

        destination = flask.session.pop('destination')
        if is_processing_fragment_encoded_response:
            # if the POST request was from the JS page handling fragment encoded responses we need to return
            # the destination URL as the response body
            return destination

        return redirect(destination)

    def _handle_error_response(self, error_response, should_redirect=False):
        if should_redirect:
            # if the current request was from the JS page handling fragment encoded responses we need to return
            # a URL for the error page to redirect to
            flask.session['error'] = error_response
            return '/' + self._redirect_uri_config.endpoint + '?error=1'
        return self._show_error_response(error_response)

    def _show_error_response(self, error_response):
        logger.error(json.dumps(error_response))
        if self._error_view:
            error = {k: error_response[k] for k in ['error', 'error_description'] if k in error_response}
            return self._error_view(**error)

        return 'Something went wrong with the authentication, please try to login again.'

    def oidc_auth(self, provider_name: str):

        if provider_name not in self._provider_configurations:
            raise ValueError(
                f"Provider name '{provider_name}' not in configured providers: {self._provider_configurations.keys()}."
            )

        def oidc_decorator(view_func):

            @functools.wraps(view_func)
            def wrapper(*args, **kwargs):

                session = UserSession(flask.session, provider_name)
                client = self.clients[session.current_provider]

                if session.should_refresh(client.session_refresh_interval_seconds):
                    logger.debug('user auth will be refreshed "silently"')
                    return self._authenticate(client, interactive=False)
                elif session.is_authenticated():
                    logger.debug('user is already authenticated')
                    return view_func(*args, **kwargs)
                else:
                    logger.debug('user not authenticated, start flow')
                    return self._authenticate(client)

            return wrapper

        return oidc_decorator

    def _logout(self, post_logout_redirect_uri):
        logger.debug('user logout')
        try:
            session = UserSession(flask.session)
        except UninitialisedSession:
            logger.info('user was already logged out, doing nothing')
            return None

        id_token_jwt = session.id_token_jwt
        client = self.clients[session.current_provider]
        session.clear()

        if client.provider_end_session_endpoint:
            flask.session['end_session_state'] = rndstr()

            end_session_request = EndSessionRequest(id_token_hint=id_token_jwt,
                                                    post_logout_redirect_uri=post_logout_redirect_uri,
                                                    state=flask.session['end_session_state'])

            logger.debug('send endsession request: %s', end_session_request.to_json())

            return redirect(end_session_request.request(client.provider_end_session_endpoint), 303)
        return None

    def oidc_logout(self, view_func):
        self._logout_views.append(view_func)

        @functools.wraps(view_func)
        def wrapper(*args, **kwargs):
            if 'state' in flask.request.args:
                # returning redirect from provider
                if flask.request.args['state'] != flask.session.pop('end_session_state', None):
                    logger.error("Got unexpected state '%s' after logout redirect.", flask.request.args['state'])
                return view_func(*args, **kwargs)

            post_logout_redirect_uri = flask.request.url
            redirect_to_provider = self._logout(post_logout_redirect_uri)
            if redirect_to_provider:
                return redirect_to_provider

            return view_func(*args, **kwargs)

        return wrapper

    def error_view(self, view_func):
        self._error_view = view_func
        return view_func

    def valid_access_token(self, force_refresh=False):
        """
        Returns a valid access token.

        1. If the current access token in the user session is valid, return that.
        2. If the current access token has expired and there is a refresh token in the user session,
           make a refresh token request and return the new access token.
        3. If the token refresh fails, either due to missing refresh token or token error response, return None.

        Args:
            force_refresh (bool): whether to perform the refresh token request even if the current access token is valid
        Returns:
            Option[str]: valid access token
        """
        try:
            session = UserSession(flask.session)
        except UninitialisedSession:
            logger.debug('user does not have an active session')
            return None

        has_expired = session.access_token_expires_at < time.time() if session.access_token_expires_at else False
        if not has_expired and not force_refresh:
            logger.debug("access token doesn't need to be refreshed")
            return session.access_token

        if not session.refresh_token:
            logger.info('no refresh token exists in the session')
            return None

        client = self.clients[session.current_provider]
        response = client.refresh_token(session.refresh_token)
        if 'error' in response:
            logger.info('failed to refresh access token: ' + json.dumps(response.to_dict()))
            return None

        access_token = response.get('access_token')
        session.update(access_token=access_token,
                       expires_in=response.get('expires_in'),
                       id_token=response['id_token'].to_dict() if 'id_token' in response else None,
                       id_token_jwt=response.get('id_token_jwt'),
                       refresh_token=response.get('refresh_token'))
        return access_token

    @staticmethod
    def _check_authorization_header(request) -> bool:
        """Look for authorization in request header.

        Parameters
        ----------
        request : flask.Request
            flask request object.

        Returns
        -------
        bool
            True if the request header contains authorization else False.
        """
        if 'Authorization' in request.headers and request.headers['Authorization'].startswith('Bearer '):
            return True
        return False

    @staticmethod
    def _parse_access_token(request) -> str:
        """Parse access token from the authorization request header.

        Parameters
        ----------
        request : flask.Request
            flask request object.

        Returns
        -------
        accept_token : str
            access token from the request header.
        """
        _, access_token = request.headers['Authorization'].split(maxsplit=1)
        return access_token

    def introspect_token(self, request, client, scopes: list = None) ->\
            Optional[TokenIntrospectionResponse]:
        """RFC 7662: Token Introspection
        The Token Introspection extension defines a mechanism for resource
        servers to obtain information about access tokens. With this spec,
        resource servers can check the validity of access tokens, and find out
        other information such as which user and which scopes are associated
        with the token.

        Parameters
        ----------
        request : flask.Request
            flask request object.
        client : flask_pyoidc.pyoidc_facade.PyoidcFacade
            PyoidcFacade object contains metadata of the provider and client.
        scopes : list, optional
            Specify scopes required by your endpoint.

        Returns
        -------
        result: TokenIntrospectionResponse or None
            If access_token is valid or None if invalid.
        """
        received_access_token = self._parse_access_token(request)
        # send token introspection request
        result = client._token_introspection_request(
            access_token=received_access_token)
        logger.debug(result)
        # Check if access_token is valid, active can be True or False
        if not result.get('active'):
            return None
        # Check if client_id is in audience claim
        if client._client.client_id not in result['aud']:
            # log the exception if client_id is not in audience and returns
            # False, you can configure audience with Identity Provider
            logger.info('Token is valid but required audience is missing.')
            return None
        # Check if the scopes associated with the access_token are the ones
        # required by the endpoint and not something else which is not
        # permitted.
        if scopes and not set(scopes).issubset(set(result['scope'])):
            logger.info('Token is valid but does not have required scopes.')
            return None
        return result

    def token_auth(self, provider_name, scopes_required: list = None):
        """Token based authorization.

        Parameters
        ----------
        provider_name : str
            Name of the provider registered with OIDCAuthorization.
        scopes_required : list, optional
            List of valid scopes associated with the endpoint.

        Raises
        ------
        Unauthorized
            flask.abort(401) if authorization field is missing.
        Forbidden
            flask.abort(403) if access token is invalid.

        Examples
        --------
        ::

            auth = OIDCAuthentication({'default': provider_config})
            @app.route('/')
            @auth.token_auth(provider_name='default')
            def index():
                ...

        You can also specify scopes required by the endpoint.

        ::

            @auth.token_auth(provider_name='default',
                             scopes_required=['read', 'write'])
        """
        def token_decorator(view_func):

            @functools.wraps(view_func)
            def wrapper(*args, **kwargs):

                client = self.clients[provider_name]
                # Check for authorization field in the request header.
                if not self._check_authorization_header(flask.request):
                    logger.info('Request header has no authorization field')
                    # Abort the request if authorization field is missing.
                    flask.abort(401)
                token_introspection_result = self.introspect_token(
                    request=flask.request, client=client,
                    scopes=scopes_required)
                if token_introspection_result:
                    logger.info('Request has valid access token.')
                    # Store token introspection info within the application
                    # context.
                    g.current_token_identity = token_introspection_result.to_dict()
                    return view_func(*args, **kwargs)
                # Forbid access if the access token is invalid.
                flask.abort(403)

            return wrapper

        return token_decorator

    def access_control(self, provider_name: str,
                       scopes_required: list = None):
        """This decorator serves dual purpose that is it can do both token
        based authorization and oidc based authentication. If your API needs
        to be accessible by either modes, use this decorator otherwise use
        either oidc_auth or token_auth.

        Parameters
        ----------
        provider_name : str
            Name of the provider registered with OIDCAuthorization.
        scopes_required : list, optional
            List of valid scopes associated with the endpoint.

        Raises
        ------
        Forbidden
            If accesss_token is invalid.

        Examples
        --------
        ::

            auth = OIDCAuthentication({'default': provider_config})
            @app.route('/')
            @auth.access_control(provider_name='default')
            def index():
                ...

        You can also specify scopes required by the endpoint:

        ::

            @auth.access_control(provider_name='default',
                                 scopes_required=['read', 'write'])
        """
        def hybrid_decorator(view_func):

            fallback_to_oidc = self.oidc_auth(provider_name)(view_func)

            @functools.wraps(view_func)
            def wrapper(*args, **kwargs):

                try:
                    # If the request header contains authorization, token_auth
                    # verifies the access_token otherwise an exception occurs
                    # and the request falls back to oidc_auth.
                    return self.token_auth(provider_name, scopes_required)(
                        view_func)(*args, **kwargs)
                # token_auth will raise the HTTPException if either
                # authorization field is missing from the request header or
                # token is invalid. If the authorization field is missing,
                # fallback to oidc.
                except Unauthorized:
                    return fallback_to_oidc(*args, **kwargs)
                # If token is present, but it's invalid, do not fall back to
                # oidc_auth. Instead, abort the request.
                except Forbidden:
                    flask.abort(403)

            return wrapper

        return hybrid_decorator



import pytest
from flask_pyoidc.redirect_uri_config import RedirectUriConfig


class TestRedirectUriConfig:
    LEGACY_CONFIG = {'SERVER_NAME': 'example.com', 'PREFERRED_URL_SCHEME': 'http'}

    def test_legacy_config_defaults(self):
        config = RedirectUriConfig.from_config(self.LEGACY_CONFIG)
        assert config.endpoint == 'redirect_uri'
        assert config.full_uri == 'http://example.com/redirect_uri'

    def test_legacy_config_endpoint(self):
        config = RedirectUriConfig.from_config({'OIDC_REDIRECT_ENDPOINT': '/foo', **self.LEGACY_CONFIG})
        assert config.endpoint == 'foo'

    def test_legacy_config_domain(self):
        config = {
            'OIDC_REDIRECT_DOMAIN': 'other.example.com:6000',  # should be preferred over SERVER_NAME
            **self.LEGACY_CONFIG
        }
        redirect_uri_config = RedirectUriConfig.from_config(config)
        assert redirect_uri_config.full_uri == 'http://other.example.com:6000/redirect_uri'

    def test_redirect_uri_config(self):
        config = {
            'OIDC_REDIRECT_URI': 'https://myexample.com:6000/callback',  # should be preferred over all other config
            'OIDC_REDIRECT_DOMAIN': 'other.example.com:6000',
            **self.LEGACY_CONFIG
        }
        redirect_uri_config = RedirectUriConfig.from_config(config)
        assert redirect_uri_config.full_uri == 'https://myexample.com:6000/callback'
        assert redirect_uri_config.endpoint == 'callback'
        assert repr(redirect_uri_config) == f'({redirect_uri_config.full_uri}, {redirect_uri_config.endpoint})'

    def test_should_raise_if_missing_all_config(self):
        with pytest.raises(ValueError) as exc_info:
            RedirectUriConfig.from_config({})
        assert 'OIDC_REDIRECT_URI' in str(exc_info.value)



import time
from urllib.parse import parse_qsl

import pytest
import responses
from oic.oic import (AccessTokenResponse, AuthorizationErrorResponse,
                     AuthorizationResponse, Grant, OpenIDSchema,
                     TokenErrorResponse)

from flask_pyoidc.provider_configuration import (ClientMetadata,
                                                 ClientRegistrationInfo,
                                                 ProviderConfiguration,
                                                 ProviderMetadata)
from flask_pyoidc.pyoidc_facade import PyoidcFacade

from .util import signed_id_token

REDIRECT_URI = 'https://rp.example.com/redirect_uri'


class TestPyoidcFacade:
    PROVIDER_BASEURL = 'https://op.example.com'
    PROVIDER_METADATA = ProviderMetadata(issuer=PROVIDER_BASEURL,
                                         authorization_endpoint=PROVIDER_BASEURL + '/auth',
                                         jwks_uri=PROVIDER_BASEURL + '/jwks')
    CLIENT_METADATA = ClientMetadata('client1', 'secret1')

    def test_registered_client_metadata_is_forwarded_to_pyoidc(self):
        config = ProviderConfiguration(provider_metadata=self.PROVIDER_METADATA, client_metadata=self.CLIENT_METADATA)
        facade = PyoidcFacade(config, REDIRECT_URI)

        expected = {
            'client_id': self.CLIENT_METADATA['client_id'],
            'client_secret': self.CLIENT_METADATA['client_secret'],
            'redirect_uris': [REDIRECT_URI],
            'token_endpoint_auth_method': 'client_secret_basic'
        }
        assert facade._client.registration_response.to_dict() == expected

    def test_no_registered_client_metadata_is_handled(self):
        config = ProviderConfiguration(provider_metadata=self.PROVIDER_METADATA,
                                       client_registration_info=ClientRegistrationInfo())
        facade = PyoidcFacade(config, REDIRECT_URI)
        assert not facade._client.registration_response

    def test_is_registered(self):
        unregistered = ProviderConfiguration(provider_metadata=self.PROVIDER_METADATA,
                                             client_registration_info=ClientRegistrationInfo())
        registered = ProviderConfiguration(provider_metadata=self.PROVIDER_METADATA,
                                           client_metadata=self.CLIENT_METADATA)
        assert PyoidcFacade(unregistered, REDIRECT_URI).is_registered() is False
        assert PyoidcFacade(registered, REDIRECT_URI).is_registered() is True

    @responses.activate
    def test_register(self):
        registration_endpoint = self.PROVIDER_BASEURL + '/register'
        redirect_uris = ['https://client.example.com/redirect']
        post_logout_redirect_uris = ['https://client.example.com/logout']
        client_registration_response = {
            'client_id': 'client1',
            'client_secret': 'secret1',
            'client_name': 'Test Client',
            'redirect_uris': redirect_uris,
            'post_logout_redirect_uris': post_logout_redirect_uris
        }
        responses.add(responses.POST, registration_endpoint, json=client_registration_response)

        provider_metadata = self.PROVIDER_METADATA.copy(registration_endpoint=registration_endpoint)
        unregistered = ProviderConfiguration(provider_metadata=provider_metadata,
                                             client_registration_info=ClientRegistrationInfo(
                                                 redirect_uris=redirect_uris,
                                                 post_logout_redirect_uris=post_logout_redirect_uris
                                             ))
        facade = PyoidcFacade(unregistered, REDIRECT_URI)
        facade.register()
        assert facade.is_registered() is True

    def test_authentication_request(self):
        extra_user_auth_params = {'foo': 'bar', 'abc': 'xyz'}
        config = ProviderConfiguration(provider_metadata=self.PROVIDER_METADATA,
                                       client_metadata=self.CLIENT_METADATA,
                                       auth_request_params=extra_user_auth_params)

        state = 'test_state'
        nonce = 'test_nonce'

        facade = PyoidcFacade(config, REDIRECT_URI)
        extra_lib_auth_params = {'foo': 'baz', 'qwe': 'rty'}
        auth_request = facade.authentication_request(state, nonce, extra_lib_auth_params)

        expected_auth_params = {
            'scope': 'openid',
            'response_type': 'code',
            'client_id': self.CLIENT_METADATA['client_id'],
            'redirect_uri': REDIRECT_URI,
            'state': state,
            'nonce': nonce
        }
        expected_auth_params.update(extra_user_auth_params)
        expected_auth_params.update(extra_lib_auth_params)
        assert auth_request.to_dict() == expected_auth_params

    def test_parse_authentication_response(self):
        facade = PyoidcFacade(ProviderConfiguration(provider_metadata=self.PROVIDER_METADATA,
                                                    client_metadata=self.CLIENT_METADATA),
                              REDIRECT_URI)
        auth_code = 'auth_code-1234'
        state = 'state-1234'
        auth_response = AuthorizationResponse(**{'state': state, 'code': auth_code})
        parsed_auth_response = facade.parse_authentication_response(auth_response.to_dict())
        assert isinstance(parsed_auth_response, AuthorizationResponse)
        assert parsed_auth_response.to_dict() == auth_response.to_dict()

    def test_parse_authentication_response_handles_error_response(self):
        facade = PyoidcFacade(ProviderConfiguration(provider_metadata=self.PROVIDER_METADATA,
                                                    client_metadata=self.CLIENT_METADATA),
                              REDIRECT_URI)
        error_response = AuthorizationErrorResponse(**{'error': 'invalid_request', 'state': 'state-1234'})
        parsed_auth_response = facade.parse_authentication_response(error_response)
        assert isinstance(parsed_auth_response, AuthorizationErrorResponse)
        assert parsed_auth_response.to_dict() == error_response.to_dict()

    @responses.activate
    def test_parse_authentication_response_preserves_id_token_jwt(self):
        facade = PyoidcFacade(ProviderConfiguration(provider_metadata=self.PROVIDER_METADATA,
                                                    client_metadata=self.CLIENT_METADATA),
                              REDIRECT_URI)
        state = 'state-1234'
        now = int(time.time())
        id_token, id_token_signing_key = signed_id_token({
            'iss': self.PROVIDER_METADATA['issuer'],
            'sub': 'test_sub',
            'aud': 'client1',
            'exp': now + 1,
            'iat': now
        })
        responses.add(responses.GET,
                      self.PROVIDER_METADATA['jwks_uri'],
                      json={'keys': [id_token_signing_key.serialize()]})
        auth_response = AuthorizationResponse(**{'state': state, 'id_token': id_token})
        parsed_auth_response = facade.parse_authentication_response(auth_response)
        assert isinstance(parsed_auth_response, AuthorizationResponse)
        assert parsed_auth_response['state'] == state
        assert parsed_auth_response['id_token_jwt'] == id_token

    @pytest.mark.parametrize('request_func, expected_token_request', [
        (
                lambda facade: facade.exchange_authorization_code('auth-code', 'test-state', {}),
                {
                    'grant_type': 'authorization_code',
                    'state': 'test-state',
                    'redirect_uri': REDIRECT_URI
                }
        ),
        (
                lambda facade: facade.refresh_token('refresh-token'),
                {
                    'grant_type': 'refresh_token',
                    'refresh_token': 'refresh-token',
                    'redirect_uri': REDIRECT_URI
                }
        )
    ])
    @responses.activate
    def test_token_request(self, request_func, expected_token_request):
        token_endpoint = self.PROVIDER_BASEURL + '/token'
        now = int(time.time())
        id_token_claims = {
            'iss': self.PROVIDER_METADATA['issuer'],
            'sub': 'test_user',
            'aud': [self.CLIENT_METADATA['client_id']],
            'exp': now + 1,
            'iat': now,
            'nonce': 'test_nonce'
        }
        id_token_jwt, id_token_signing_key = signed_id_token(id_token_claims)
        token_response = AccessTokenResponse(access_token='test_access_token',
                                             refresh_token='refresh-token',
                                             token_type='Bearer',
                                             id_token=id_token_jwt,
                                             expires_in=now + 1)

        responses.add(responses.POST, token_endpoint, json=token_response.to_dict())

        provider_metadata = self.PROVIDER_METADATA.copy(token_endpoint=token_endpoint)
        facade = PyoidcFacade(ProviderConfiguration(provider_metadata=provider_metadata,
                                                    client_metadata=self.CLIENT_METADATA),
                              REDIRECT_URI)
        grant = Grant(resp=token_response)
        grant.grant_expiration_time = now + grant.exp_in
        facade._client.grant = {'test-state': grant}

        responses.add(responses.GET,
                      self.PROVIDER_METADATA['jwks_uri'],
                      json={'keys': [id_token_signing_key.serialize()]})
        token_response = request_func(facade)

        assert isinstance(token_response, AccessTokenResponse)
        expected_token_response = token_response.to_dict()
        expected_token_response['id_token'] = id_token_claims
        expected_token_response['id_token_jwt'] = id_token_jwt
        assert token_response.to_dict() == expected_token_response

        token_request = dict(parse_qsl(responses.calls[0].request.body))
        assert token_request == expected_token_request

    @responses.activate
    def test_token_request_handles_error_response(self):
        token_endpoint = self.PROVIDER_BASEURL + '/token'
        token_response = TokenErrorResponse(error='invalid_request', error_description='test error description')
        responses.add(responses.POST, token_endpoint, json=token_response.to_dict(), status=400)

        provider_metadata = self.PROVIDER_METADATA.copy(token_endpoint=token_endpoint)
        facade = PyoidcFacade(ProviderConfiguration(provider_metadata=provider_metadata,
                                                    client_metadata=self.CLIENT_METADATA),
                              REDIRECT_URI)
        state = 'test-state'
        grant = Grant()
        grant.grant_expiration_time = int(time.time()) + grant.exp_in
        facade._client.grant = {state: grant}
        assert facade.exchange_authorization_code('1234', state, {}) == token_response

    def test_token_request_handles_missing_provider_token_endpoint(self):
        facade = PyoidcFacade(ProviderConfiguration(provider_metadata=self.PROVIDER_METADATA,
                                                    client_metadata=self.CLIENT_METADATA),
                              REDIRECT_URI)
        assert facade.exchange_authorization_code(None, None, {}) is None

    @pytest.mark.parametrize('userinfo_http_method', [
        'GET',
        'POST'
    ])
    @responses.activate
    def test_configurable_userinfo_endpoint_method_is_used(self, userinfo_http_method):
        userinfo_endpoint = self.PROVIDER_BASEURL + '/userinfo'
        userinfo_response = OpenIDSchema(sub='user1')
        responses.add(userinfo_http_method, userinfo_endpoint, json=userinfo_response.to_dict())

        provider_metadata = self.PROVIDER_METADATA.copy(userinfo_endpoint=userinfo_endpoint)
        facade = PyoidcFacade(ProviderConfiguration(provider_metadata=provider_metadata,
                                                    client_metadata=self.CLIENT_METADATA,
                                                    userinfo_http_method=userinfo_http_method),
                              REDIRECT_URI)
        assert facade.userinfo_request('test_token') == userinfo_response

    def test_no_userinfo_request_is_made_if_no_userinfo_http_method_is_configured(self):
        facade = PyoidcFacade(ProviderConfiguration(provider_metadata=self.PROVIDER_METADATA,
                                                    client_metadata=self.CLIENT_METADATA,
                                                    userinfo_http_method=None),
                              REDIRECT_URI)
        assert facade.userinfo_request('test_token') is None

    def test_no_userinfo_request_is_made_if_no_userinfo_endpoint_is_configured(self):
        facade = PyoidcFacade(ProviderConfiguration(provider_metadata=self.PROVIDER_METADATA,
                                                    client_metadata=self.CLIENT_METADATA),
                              REDIRECT_URI)
        assert facade.userinfo_request('test_token') is None

    def test_no_userinfo_request_is_made_if_no_access_token(self):
        provider_metadata = self.PROVIDER_METADATA.copy(userinfo_endpoint=self.PROVIDER_BASEURL + '/userinfo')
        facade = PyoidcFacade(ProviderConfiguration(provider_metadata=provider_metadata,
                                                    client_metadata=self.CLIENT_METADATA),
                              REDIRECT_URI)
        assert facade.userinfo_request(None) is None

    @responses.activate
    @pytest.mark.parametrize('scope, extra_args',
                             [(None, {}),
                              (['read', 'write'],
                               {'audience': ['client_id1, client_id2']})
                              ])
    def test_client_credentials_grant(self, scope, extra_args):
        token_endpoint = f'{self.PROVIDER_BASEURL}/token'
        provider_metadata = self.PROVIDER_METADATA.copy(
            token_endpoint=token_endpoint)
        facade = PyoidcFacade(
            ProviderConfiguration(
                provider_metadata=provider_metadata,
                client_metadata=self.CLIENT_METADATA),
            REDIRECT_URI)
        client_credentials_grant_response = {
            'access_token': 'access_token',
            'expires_in': 60,
            'not-before-policy': 0,
            'refresh_expires_in': 0,
            'scope': 'read write',
            'token_type': 'Bearer'}
        responses.add(responses.POST, token_endpoint,
                      json=client_credentials_grant_response)
        assert client_credentials_grant_response == facade.client_credentials_grant(
            scope=scope, **extra_args).to_dict()

    def test_post_logout_redirect_uris(self):
        post_logout_redirect_uris = ['https://client.example.com/logout']
        client_metadata = self.CLIENT_METADATA.copy(
            post_logout_redirect_uris=post_logout_redirect_uris)
        facade = PyoidcFacade(ProviderConfiguration(provider_metadata=self.PROVIDER_METADATA,
                                                    client_metadata=client_metadata),
                              REDIRECT_URI)
        assert facade.post_logout_redirect_uris == post_logout_redirect_uris



import time


class UninitialisedSession(Exception):
    pass


class UserSession:
    """Session object for user login state.

    Wraps comparison of times necessary for session handling.
    """

    KEYS = [
        'access_token',
        'access_token_expires_at',
        'current_provider',
        'id_token',
        'id_token_jwt',
        'last_authenticated',
        'last_session_refresh',
        'userinfo',
        'refresh_token'
    ]

    def __init__(self, session_storage, provider_name=None):
        self._session_storage = session_storage
        if 'current_provider' not in self._session_storage and not provider_name:
            raise UninitialisedSession("Trying to pick-up uninitialised session without specifying 'provider_name'")

        if provider_name:
            if 'current_provider' in self._session_storage and \
                    provider_name != self._session_storage['current_provider']:
                # provider has changed, initialise new session
                self.clear()

            self._session_storage['current_provider'] = provider_name

    def is_authenticated(self):
        """
        flask_session is empty when the session hasn't been initialised or has expired.
        Thus checking for existence of any item is enough to determine if we're authenticated.
        """

        return self._session_storage.get('last_authenticated') is not None

    def should_refresh(self, refresh_interval_seconds=None):
        return refresh_interval_seconds is not None and \
               self._session_storage.get('last_session_refresh') is not None and \
               self._refresh_time(refresh_interval_seconds) < time.time()

    def _refresh_time(self, refresh_interval_seconds):
        last = self._session_storage.get('last_session_refresh', 0)
        return last + refresh_interval_seconds

    def update(self, *,
               access_token=None,
               expires_in=None,
               id_token=None,
               id_token_jwt=None,
               userinfo=None,
               refresh_token=None):
        """
        Args:
            access_token (str)
            expires_in (int)
            id_token (Mapping[str, str])
            id_token_jwt (str)
            userinfo (Mapping[str, str])
        """

        def set_if_defined(session_key, value):
            if value:
                self._session_storage[session_key] = value

        now = int(time.time())
        auth_time = now
        if id_token:
            auth_time = id_token.get('auth_time', auth_time)

        self._session_storage['last_authenticated'] = auth_time
        self._session_storage['last_session_refresh'] = now
        set_if_defined('access_token', access_token)
        set_if_defined('access_token_expires_at', now + expires_in if expires_in else None)
        set_if_defined('id_token', id_token)
        set_if_defined('id_token_jwt', id_token_jwt)
        set_if_defined('userinfo', userinfo)
        set_if_defined('refresh_token', refresh_token)

    def clear(self):
        for key in self.KEYS:
            self._session_storage.pop(key, None)

    @property
    def access_token(self):
        return self._session_storage.get('access_token')

    @property
    def access_token_expires_at(self):
        return self._session_storage.get('access_token_expires_at')

    @property
    def refresh_token(self):
        return self._session_storage.get('refresh_token')

    @property
    def id_token(self):
        return self._session_storage.get('id_token')

    @property
    def id_token_jwt(self):
        return self._session_storage.get('id_token_jwt')

    @property
    def userinfo(self):
        return self._session_storage.get('userinfo')

    @property
    def current_provider(self):
        return self._session_storage.get('current_provider')

    @property
    def last_authenticated(self):
        return self._session_storage.get('last_authenticated')
